import array
import copy
import math
import xml.etree.ElementTree as ET

import time

from urh.signalprocessing.Encoding import Encoding
from urh.signalprocessing.FieldType import FieldType
from urh.signalprocessing.MessageType import MessageType
from urh.signalprocessing.Participant import Participant
from urh.signalprocessing.ProtocoLabel import ProtocolLabel
from urh.util.Formatter import Formatter
from urh.util.Logger import logger


class Message(object):
    """
    A protocol message is a single line of a protocol.
    """

    __slots__ = [
        "__plain_bits",
        "__bit_alignments",
        "pause",
        "modulator_index",
        "rssi",
        "participant",
        "message_type",
        "absolute_time",
        "relative_time",
        "__decoder",
        "align_labels",
        "decoding_state",
        "timestamp",
        "fuzz_created",
        "__decoded_bits",
        "__encoded_bits",
        "decoding_errors",
        "samples_per_symbol",
        "bit_sample_pos",
        "alignment_offset",
        "bits_per_symbol",
    ]

    def __init__(
        self,
        plain_bits,
        pause: int,
        message_type: MessageType,
        rssi=0,
        modulator_index=0,
        decoder=None,
        fuzz_created=False,
        bit_sample_pos=None,
        samples_per_symbol=100,
        participant=None,
        bits_per_symbol=1,
        timestamp=0,
    ):
        """

        :param pause: pause AFTER the message in samples
        :type plain_bits: list[bool|int]
        :type decoder: Encoding
        :type bit_alignment_positions: list of int
        :param bit_alignment_positions: Für Ausrichtung der Hex Darstellung (Leere Liste für Standardverhalten)
        :param samples_per_symbol: Für Übernahme der Bitlänge in Modulator Dialog
        :param fuzz_created: message was created through fuzzing
        :return:
        """
        self.__plain_bits = array.array("B", plain_bits)
        self.pause = pause
        self.modulator_index = modulator_index
        self.rssi = rssi
        self.participant = participant  # type: Participant
        self.message_type = message_type  # type: MessageType

        if timestamp == 0:
            self.timestamp = time.time()
        else:
            # Caller passed specific timestamp for this message
            self.timestamp = timestamp

        self.absolute_time = 0  # set in Compare Frame
        self.relative_time = 0  # set in Compare Frame

        self.__decoder = (
            decoder if decoder else Encoding(["Non Return To Zero (NRZ)"])
        )  # type: Encoding

        self.align_labels = True
        self.fuzz_created = fuzz_created

        self.alignment_offset = 0

        self.__decoded_bits = None
        self.__encoded_bits = None
        self.__bit_alignments = []
        self.decoding_errors = 0
        self.decoding_state = Encoding.ErrorState.SUCCESS

        self.samples_per_symbol = samples_per_symbol  # to take over in modulator
        self.bits_per_symbol = bits_per_symbol  # to take over in generator tab (default modulator settings)

        if bit_sample_pos is None:
            self.bit_sample_pos = array.array("L", [])
        else:
            self.bit_sample_pos = bit_sample_pos
            """
            :param bit_sample_pos: Position of samples for each bit. Last position is pause so last bit is on pos -2.
            :type  bit_sample_pos: array.array
            """

    @property
    def plain_bits(self):
        """

        :rtype: array.array
        """
        return self.__plain_bits

    @plain_bits.setter
    def plain_bits(self, value: list):
        self.__plain_bits = array.array("B", value)
        self.clear_decoded_bits()
        self.clear_encoded_bits()

    @property
    def active_fuzzing_labels(self):
        return [lbl for lbl in self.message_type if lbl.active_fuzzing]

    @property
    def exclude_from_decoding_labels(self):
        return [lbl for lbl in self.message_type if not lbl.apply_decoding]

    def __getitem__(self, index: int):
        return self.plain_bits[index]

    def __setitem__(self, index: int, value):
        """

        :type value: bool
        """
        self.plain_bits[index] = value
        self.clear_decoded_bits()
        self.clear_encoded_bits()

    def __add__(self, other):
        return self.__plain_bits + other.__plain_bits

    def _remove_labels_for_range(self, index, instant_remove=True):
        if isinstance(index, int):
            index = slice(index, index + 1, 1)

        assert isinstance(index, slice)

        start = index.start if index.start is not None else 0
        stop = index.stop
        step = index.step if index.step is not None else 1

        removed_labels = []

        for lbl in self.message_type:  # type: ProtocolLabel
            if (
                (start <= lbl.start and stop >= lbl.end)
                or start <= lbl.start <= stop
                or (start >= lbl.start and stop <= lbl.end)
                or lbl.start <= start < lbl.end
            ):
                if instant_remove:
                    self.message_type.remove(lbl)
                removed_labels.append(lbl)

            elif stop - 1 < lbl.start:
                number_elements = len(range(start, stop, step))
                l_cpy = lbl.get_copy()
                l_cpy.start -= number_elements
                l_cpy.end -= number_elements

                if instant_remove:
                    self.message_type.remove(lbl)
                    self.message_type.append(l_cpy)

        return removed_labels

    def __delitem__(self, index):
        self._remove_labels_for_range(index)
        del self.plain_bits[index]
        self.clear_decoded_bits()
        self.clear_encoded_bits()

    def __str__(self):
        return self.bits2string(self.plain_bits)

    def delete_range_without_label_range_update(self, start: int, end: int):
        del self.plain_bits[start:end]
        self.clear_decoded_bits()
        self.clear_encoded_bits()

    def get_byte_length(self, decoded=True) -> int:
        """
        Return the length of this message in byte.

        """
        end = len(self.decoded_bits) if decoded else len(self.__plain_bits)
        end = self.convert_index(end, 0, 2, decoded=decoded)[0]
        return int(end)

    def bits2string(self, bits: array.array) -> str:
        return "".join(map(str, bits))

    def __len__(self):
        return len(self.plain_bits)

    def insert(self, index: int, item: bool):
        self.plain_bits.insert(index, item)
        self.clear_decoded_bits()
        self.clear_encoded_bits()

    @property
    def decoder(self) -> Encoding:
        return self.__decoder

    @decoder.setter
    def decoder(self, val: Encoding):
        self.__decoder = val
        self.clear_decoded_bits()
        self.clear_encoded_bits()
        self.decoding_errors, self.decoding_state = self.decoder.analyze(
            self.plain_bits
        )

    @property
    def encoded_bits(self):
        """

        :rtype: array.array
        """
        if self.__encoded_bits is None:
            self.__encoded_bits = array.array("B", [])
            start = 0
            encode = self.decoder.encode
            bits = self.plain_bits

            for label in self.exclude_from_decoding_labels:
                self.__encoded_bits.extend(encode(bits[start : label.start]))
                start = label.start if label.start > start else start  # Overlapping
                self.__encoded_bits.extend(bits[start : label.end])
                start = label.end if label.end > start else start  # Overlapping

            self.__encoded_bits.extend(encode(bits[start:]))
        return self.__encoded_bits

    @property
    def encoded_bits_str(self) -> str:
        return self.bits2string(self.encoded_bits)

    @property
    def decoded_bits(self) -> array.array:
        if self.__decoded_bits is None:
            self.__decoded_bits = array.array("B", [])
            start = 0
            code = self.decoder.code  # 0 = decoded, 1 = analyzed
            # decode = self.decoder.decode
            # analyze = self.decoder.analyze
            bits = self.plain_bits
            self.decoding_errors = 0
            states = set()
            self.decoding_state = self.decoder.ErrorState.SUCCESS
            for label in self.exclude_from_decoding_labels:
                decoded, errors, state = code(True, bits[start : label.start])
                states.add(state)
                self.__decoded_bits.extend(decoded)
                self.decoding_errors += errors

                if label.start == -1 or label.end == -1:
                    label.start = len(self.__decoded_bits)
                    label.end = label.start + (label.end - label.start)

                start = (
                    label.start if label.start > start else start
                )  # Überlappende Labels -.-
                self.__decoded_bits.extend(bits[start : label.end])
                start = (
                    label.end if label.end > start else start
                )  # Überlappende Labels FFS >.<

            decoded, errors, state = code(True, bits[start:])
            states.add(state)
            self.__decoded_bits.extend(decoded)
            self.decoding_errors += errors

            states.discard(self.decoder.ErrorState.SUCCESS)
            if len(states) > 0:
                self.decoding_state = sorted(states)[0]

        return self.__decoded_bits

    @decoded_bits.setter
    def decoded_bits(self, val):
        self.__decoded_bits = array.array("B", val)

    @property
    def decoded_bits_str(self) -> str:
        return self.bits2string(self.decoded_bits)

    @property
    def plain_bits_str(self) -> str:
        return str(self)

    @property
    def decoded_bits_buffer(self) -> bytes:
        return self.decoded_bits.tobytes()

    @property
    def plain_hex_array(self) -> array.array:
        padded_bitchains = self.split(decode=False)
        return self.__bit_chains_to_hex(padded_bitchains)

    @property
    def plain_hex_str(self) -> str:
        return "".join(map(lambda h: "{0:x}".format(h), self.plain_hex_array))

    @property
    def plain_ascii_array(self) -> array.array:
        padded_bitchains = self.split(decode=False)
        return self.__bit_chains_to_ascii(padded_bitchains)

    @property
    def plain_ascii_str(self) -> str:
        return "".join(map(chr, self.plain_ascii_array))

    @property
    def decoded_hex_array(self) -> array.array:
        padded_bitchains = self.split()
        return self.__bit_chains_to_hex(padded_bitchains)

    @property
    def decoded_hex_str(self) -> str:
        return "".join(map(lambda h: "{0:x}".format(h), self.decoded_hex_array))

    @property
    def decoded_ascii_array(self) -> array.array:
        padded_bitchains = self.split()
        return self.__bit_chains_to_ascii(padded_bitchains)

    @property
    def decoded_ascii_str(self) -> str:
        return "".join(map(chr, self.decoded_ascii_array))

    @property
    def decoded_ascii_buffer(self) -> bytes:
        return self.decoded_ascii_array.tobytes()

    def __get_bit_range_from_hex_or_ascii_index(
        self, from_index: int, decoded: bool, is_hex: bool
    ) -> tuple:
        bits = self.decoded_bits if decoded else self.plain_bits
        factor = 4 if is_hex else 8
        for i in range(len(bits)):
            if (
                self.__get_hex_ascii_index_from_bit_index(i, to_hex=is_hex)[0]
                == from_index
            ):
                return i, i + factor - 1

        return factor * from_index, factor * (from_index + 1) - 1

    def __get_hex_ascii_index_from_bit_index(
        self, bit_index: int, to_hex: bool
    ) -> tuple:
        factor = 4 if to_hex else 8
        result = 0

        last_alignment = 0
        for ba in self.__bit_alignments:
            if ba <= bit_index:
                result += math.ceil((ba - last_alignment) / factor)
                last_alignment = ba
            else:
                break

        result += math.floor((bit_index - last_alignment) / factor)

        return result, result

    def convert_index(self, index: int, from_view: int, to_view: int, decoded: bool):
        if to_view == from_view:
            return index, index

        if to_view == 0:
            return self.__get_bit_range_from_hex_or_ascii_index(
                index, decoded, is_hex=from_view == 1
            )
        if to_view == 1:
            if from_view == 0:
                return self.__get_hex_ascii_index_from_bit_index(index, to_hex=True)
            elif from_view == 2:
                bi = self.__get_bit_range_from_hex_or_ascii_index(
                    index, decoded, is_hex=True
                )[0]
                return self.__get_hex_ascii_index_from_bit_index(bi, to_hex=False)
        elif to_view == 2:
            if from_view == 0:
                return self.__get_hex_ascii_index_from_bit_index(index, to_hex=False)
            elif from_view == 1:
                bi = self.__get_bit_range_from_hex_or_ascii_index(
                    index, decoded, is_hex=False
                )[0]
                return self.__get_hex_ascii_index_from_bit_index(bi, to_hex=True)
        else:
            raise NotImplementedError("Only Three View Types (Bit/Hex/ASCII)")

    def convert_range(
        self, index1: int, index2: int, from_view: int, to_view: int, decoded: bool
    ):
        start = self.convert_index(index1, from_view, to_view, decoded)[0]
        end = self.convert_index(index2, from_view, to_view, decoded)[1]

        try:
            return int(start), int(math.ceil(end))
        except TypeError:
            return 0, 0

    def get_duration(self, sample_rate: int) -> float:
        if len(self.bit_sample_pos) < 2:
            raise ValueError("Not enough bit samples for calculating duration")

        return (self.bit_sample_pos[-1] - self.bit_sample_pos[0]) / sample_rate

    def get_src_address_from_data(self, decoded=True):
        """
        Return the SRC address of a message if SRC_ADDRESS label is present in message type of the message
        Return None otherwise

        :param decoded:
        :return:
        """
        src_address_label = next(
            (
                lbl
                for lbl in self.message_type
                if lbl.field_type
                and lbl.field_type.function == FieldType.Function.SRC_ADDRESS
            ),
            None,
        )
        if src_address_label:
            start, end = self.get_label_range(src_address_label, view=1, decode=decoded)
            if decoded:
                src_address = self.decoded_hex_str[start:end]
            else:
                src_address = self.plain_hex_str[start:end]
        else:
            src_address = None

        return src_address

    @staticmethod
    def __bit_chains_to_hex(bit_chains) -> array.array:
        """

        :type bit_chains: list of array.array
        :return:
        """
        result = array.array("B", [])
        for bc in bit_chains:
            bc += array.array("B", [0] * ((4 - len(bc) % 4) % 4))  # pad hex view
            result.extend(
                (8 * bc[i] + 4 * bc[i + 1] + 2 * bc[i + 2] + bc[i + 3])
                for i in range(0, len(bc), 4)
            )

        return result

    @staticmethod
    def __bit_chains_to_ascii(bit_chains) -> array.array:
        """

        :type bit_chains: list of array.array
        :return:
        """
        result = array.array("B", [])
        for bc in bit_chains:
            bc += array.array("B", [0] * ((8 - len(bc) % 8) % 8))  # pad ascii view
            result.extend(
                (
                    128 * bc[i]
                    + 64 * bc[i + 1]
                    + 32 * bc[i + 2]
                    + 16 * bc[i + 3]
                    + 8 * bc[i + 4]
                    + 4 * bc[i + 5]
                    + 2 * bc[i + 6]
                    + bc[i + 7]
                )
                for i in range(0, len(bc), 8)
            )
        return result

    def split(self, decode=True):
        """
        Für das Bit-Alignment (neu Ausrichten von Hex, ASCII-View)

        :rtype: list of array.array
        """
        start = 0
        result = []
        message = self.decoded_bits if decode else self.plain_bits
        bit_alignments = set()
        if self.align_labels:
            for l in self.message_type:
                bit_alignments.add(l.start)
                bit_alignments.add(l.end)

        self.__bit_alignments = sorted(bit_alignments)

        for pos in self.__bit_alignments:
            result.append(message[start:pos])
            start = pos

        result.append(message[start:])
        return result

    def view_to_string(
        self, view: int, decoded: bool, show_pauses=True, sample_rate: float = None
    ) -> str:
        """

        :param view: 0 - Bits ## 1 - Hex ## 2 - ASCII
        """
        if view == 0:
            proto = self.decoded_bits_str if decoded else self.plain_bits_str
        elif view == 1:
            proto = self.decoded_hex_str if decoded else self.plain_hex_str
        elif view == 2:
            proto = self.decoded_ascii_str if decoded else self.plain_ascii_str
        else:
            return None

        if show_pauses:
            return "%s %s" % (proto, self.get_pause_str(sample_rate))
        else:
            return proto

    def get_pause_str(self, sample_rate):
        if sample_rate:
            return " [<b>Pause:</b> %s]" % (
                Formatter.science_time(self.pause / sample_rate)
            )
        else:
            return " [<b>Pause:</b> %d samples]" % (self.pause)

    def clear_decoded_bits(self):
        self.__decoded_bits = None

    def clear_encoded_bits(self):
        self.__encoded_bits = None

    @staticmethod
    def from_plain_bits_str(bits, pause=0):
        plain_bits = list(map(int, bits))
        return Message(
            plain_bits=plain_bits, pause=pause, message_type=MessageType("none")
        )

    @staticmethod
    def from_plain_hex_str(hex_str, pause=0):
        lut = {"{0:x}".format(i): "{0:04b}".format(i) for i in range(16)}
        bits = "".join((lut[h] for h in hex_str))
        return Message.from_plain_bits_str(bits, pause)

    def to_xml(
        self, decoders=None, include_message_type=False, write_bits=False
    ) -> ET.Element:
        root = ET.Element("message")
        root.set("message_type_id", self.message_type.id)
        root.set("modulator_index", str(self.modulator_index))
        root.set("pause", str(self.pause))
        root.set("timestamp", str(self.timestamp))

        if write_bits:
            root.set("bits", self.plain_bits_str)

        if decoders:
            try:
                decoding_index = decoders.index(self.decoder)
            except ValueError:
                logger.warning(
                    "Failed to find '{}' in list of decodings".format(self.decoder.name)
                )
                decoding_index = 0
            root.set("decoding_index", str(decoding_index))
        if self.participant is not None:
            root.set("participant_id", self.participant.id)
        if include_message_type:
            root.append(self.message_type.to_xml())
        return root

    def from_xml(
        self, tag: ET.Element, participants, decoders=None, message_types=None
    ):
        timestamp = tag.get("timestamp", None)
        if timestamp:
            self.timestamp = float(timestamp)

        part_id = tag.get("participant_id", None)
        message_type_id = tag.get("message_type_id", None)
        self.modulator_index = int(tag.get("modulator_index", self.modulator_index))
        self.pause = int(tag.get("pause", self.pause))
        decoding_index = tag.get("decoding_index", None)
        if decoding_index and decoders is not None:
            try:
                self.decoder = decoders[int(decoding_index)]
            except IndexError:
                pass

        if part_id:
            self.participant = Participant.find_matching(part_id, participants)
            if self.participant is None:
                logger.warning(
                    "No participant matched the id {0} from xml".format(part_id)
                )

        if message_type_id and message_types:
            for message_type in message_types:
                if message_type.id == message_type_id:
                    self.message_type = message_type
                    break

        message_type_tag = tag.find("message_type")
        if message_type_tag:
            self.message_type = MessageType.from_xml(message_type_tag)

    @classmethod
    def new_from_xml(
        cls, tag: ET.Element, participants, decoders=None, message_types=None
    ):
        assert "bits" in tag.attrib
        result = cls.from_plain_bits_str(bits=tag.get("bits"))
        result.from_xml(
            tag, participants, decoders=decoders, message_types=message_types
        )
        return result

    def get_label_range(
        self, lbl: ProtocolLabel, view: int, decode: bool, consider_alignment=False
    ):
        a = self.alignment_offset if consider_alignment else 0
        start = self.convert_index(
            index=lbl.start + a, from_view=0, to_view=view, decoded=decode
        )[0]
        end = self.convert_index(
            index=lbl.end + a, from_view=0, to_view=view, decoded=decode
        )[1]
        return int(start), int(end)



import copy

from PyQt5.QtWidgets import QUndoCommand

from urh.signalprocessing.ProtocolAnalyzer import ProtocolAnalyzer


class DeleteBitsAndPauses(QUndoCommand):
    def __init__(
        self,
        proto_analyzer: ProtocolAnalyzer,
        start_message: int,
        end_message: int,
        start: int,
        end: int,
        view: int,
        decoded: bool,
        subprotos=None,
        update_label_ranges=True,
    ):
        super().__init__()

        self.sub_protocols = (
            [] if subprotos is None else subprotos
        )  # type: list[ProtocolAnalyzer]
        self.view = view
        self.end = end
        self.start = start
        self.end_message = end_message
        self.start_message = start_message
        self.proto_analyzer = proto_analyzer
        self.decoded = decoded
        self.saved_messages = []
        self.removed_message_indices = []
        self.sub_protocol_history = {}  # for CFC
        self.update_label_ranges = update_label_ranges
        for sub_protocol in self.sub_protocols:
            self.sub_protocol_history[sub_protocol] = sub_protocol.messages

        self.setText("Delete")

    def redo(self):
        self.saved_messages = copy.deepcopy(
            self.proto_analyzer.messages[self.start_message : self.end_message + 1]
        )
        self.removed_message_indices = self.proto_analyzer.delete_messages(
            self.start_message,
            self.end_message,
            self.start,
            self.end,
            self.view,
            self.decoded,
            self.update_label_ranges,
        )

    def undo(self):
        for i in reversed(range(self.start_message, self.end_message + 1)):
            if i in self.removed_message_indices:
                self.proto_analyzer.messages.insert(
                    i, self.saved_messages[i - self.start_message]
                )
            else:
                try:
                    self.proto_analyzer.messages[i] = self.saved_messages[
                        i - self.start_message
                    ]
                except IndexError:
                    self.proto_analyzer.messages.append(
                        self.saved_messages[i - self.start_message]
                    )

        for sub_protocol in self.sub_protocol_history.keys():
            sub_protocol.messages = self.sub_protocol_history[sub_protocol]

        self.saved_messages.clear()
        self.removed_message_indices.clear()



import math
from collections import defaultdict

import numpy as np

from urh.awre.CommonRange import CommonRange
from urh.awre.engines.Engine import Engine
from urh.cythonext import util


class LengthEngine(Engine):
    def __init__(self, bitvectors, already_labeled=None):
        """

        :type bitvectors: list of np.ndarray
        :param bitvectors: bitvectors behind the synchronization
        """
        self.bitvectors = bitvectors
        self.already_labeled = [] if already_labeled is None else already_labeled

    def find(self, n_gram_length=8, minimum_score=0.1):
        # Consider the n_gram_length
        bitvectors_by_n_gram_length = defaultdict(list)
        for i, bitvector in enumerate(self.bitvectors):
            bin_num = int(math.ceil(len(bitvector) / n_gram_length))
            bitvectors_by_n_gram_length[bin_num].append(i)

        common_ranges_by_length = self.find_common_ranges_by_cluster(
            self.bitvectors, bitvectors_by_n_gram_length, alpha=0.7
        )

        for length, ranges in common_ranges_by_length.items():
            common_ranges_by_length[length] = self.ignore_already_labeled(
                ranges, self.already_labeled
            )

        self.filter_common_ranges(common_ranges_by_length)
        self._debug("Common Ranges:", common_ranges_by_length)

        scored_ranges = self.score_ranges(common_ranges_by_length, n_gram_length)
        self._debug("Scored Ranges", scored_ranges)

        # Take the ranges with highest score per cluster if it's score surpasses the minimum score
        high_scores_by_length = self.choose_high_scored_ranges(
            scored_ranges, bitvectors_by_n_gram_length, minimum_score
        )
        self._debug("Highscored Ranges", high_scores_by_length)
        return high_scores_by_length.values()

    @staticmethod
    def filter_common_ranges(common_ranges_by_length: dict):
        """
        Ranges must be common along length clusters
        but their values must differ, so now we rule out all ranges that are
          1. common across clusters AND
          2. have same value

        :return:
        """

        ranges = [r for rng in common_ranges_by_length.values() for r in rng]
        for rng in ranges:
            count = len(
                [
                    r
                    for r in ranges
                    if rng.start == r.start
                    and rng.length == r.length
                    and rng.value.tobytes() == r.value.tobytes()
                ]
            )
            if count < 2:
                continue

            for length in common_ranges_by_length:
                try:
                    common_ranges_by_length[length].remove(rng)
                except ValueError:
                    pass

    @staticmethod
    def score_ranges(common_ranges_by_length: dict, n_gram_length: int):
        """
        Calculate score for the common ranges

        :param common_ranges_by_length:
        :param n_gram_length:
        :return:
        """

        # The window length must be smaller than common range's length
        # and is something like 8 in case of on 8 bit integer.
        # We make this generic so e.g. 4 bit integers are supported as well
        if n_gram_length == 8:
            window_lengths = [8, 16, 32, 64]
        else:
            window_lengths = [n_gram_length * i for i in range(1, 5)]

        scored_ranges = dict()
        for length in common_ranges_by_length:
            scored_ranges[length] = dict()
            for window_length in window_lengths:
                scored_ranges[length][window_length] = []

        byteorders = ["big", "little"] if n_gram_length == 8 else ["big"]
        for window_length in window_lengths:
            for length, common_ranges in common_ranges_by_length.items():
                for common_range in filter(
                    lambda cr: cr.length >= window_length, common_ranges
                ):
                    bits = common_range.value
                    rng_byte_order = "big"

                    max_score = max_start = -1
                    for start in range(0, len(bits) + 1 - window_length, n_gram_length):
                        for byteorder in byteorders:
                            score = LengthEngine.score_bits(
                                bits[start : start + window_length],
                                length,
                                position=start,
                                byteorder=byteorder,
                            )

                            if score > max_score:
                                max_score = score
                                max_start = start
                                rng_byte_order = byteorder

                    rng = CommonRange(
                        common_range.start + max_start,
                        window_length,
                        common_range.value[max_start : max_start + window_length],
                        score=max_score,
                        field_type="length",
                        message_indices=common_range.message_indices,
                        range_type=common_range.range_type,
                        byte_order=rng_byte_order,
                    )
                    scored_ranges[length][window_length].append(rng)

        return scored_ranges

    def choose_high_scored_ranges(
        self,
        scored_ranges: dict,
        bitvectors_by_n_gram_length: dict,
        minimum_score: float,
    ):
        # Set for every window length the highest scored range as candidate
        possible_window_lengths = defaultdict(int)
        for length, ranges_by_window_length in scored_ranges.items():
            for window_length, ranges in ranges_by_window_length.items():
                try:
                    ranges_by_window_length[window_length] = max(
                        filter(lambda x: x.score >= minimum_score, ranges),
                        key=lambda x: x.score,
                    )
                    possible_window_lengths[window_length] += 1
                except ValueError:
                    ranges_by_window_length[window_length] = None

        try:
            # Choose window length -> window length that has a result most often and choose greater on tie
            chosen_window_length = max(
                possible_window_lengths, key=lambda x: (possible_window_lengths[x], x)
            )
        except ValueError:
            return dict()

        high_scores_by_length = dict()

        # Choose all ranges with highest score per cluster if score surpasses the minimum score
        for length, ranges_by_window_length in scored_ranges.items():
            try:
                if ranges_by_window_length[chosen_window_length]:
                    high_scores_by_length[length] = ranges_by_window_length[
                        chosen_window_length
                    ]
            except KeyError:
                continue

        # If there are length clusters with only one message see if we can assign a range from other clusters
        for length, msg_indices in bitvectors_by_n_gram_length.items():
            if len(msg_indices) != 1:
                continue

            msg_index = msg_indices[0]
            bitvector = self.bitvectors[msg_index]
            max_score, best_match = 0, None

            for rng in high_scores_by_length.values():
                bits = bitvector[rng.start : rng.end + 1]
                if len(bits) > 0:
                    score = self.score_bits(bits, length, rng.start)
                    if score > max_score:
                        best_match, max_score = rng, score

            if best_match is not None:
                high_scores_by_length[length] = CommonRange(
                    best_match.start,
                    best_match.length,
                    value=bitvector[best_match.start : best_match.end + 1],
                    score=max_score,
                    field_type="length",
                    message_indices={msg_index},
                    range_type="bit",
                )

        return high_scores_by_length

    @staticmethod
    def score_bits(
        bits: np.ndarray, target_length: int, position: int, byteorder="big"
    ):
        value = util.bit_array_to_number(bits, len(bits))
        if byteorder == "little":
            if len(bits) > 8 and len(bits) % 8 == 0:
                n = len(bits) // 8
                value = int.from_bytes(
                    value.to_bytes(n, byteorder="big"), byteorder="little", signed=False
                )

        # Length field should be at front, so we give lower scores for large starts
        f = 1 / (1 + 0.25 * position)

        return f * LengthEngine.gauss(value, target_length)

    @staticmethod
    def gauss(x, mu, sigma=2):
        return np.exp(-0.5 * np.power((x - mu) / sigma, 2))



import math

from PyQt5.QtCore import Qt, pyqtSlot
from PyQt5.QtWidgets import QDialog, QTableWidgetItem

from urh.ui.delegates.KillerSpinBoxDelegate import KillerSpinBoxDelegate
from urh.ui.delegates.SpinBoxDelegate import SpinBoxDelegate
from urh.ui.ui_modulation_parameters_dialog import Ui_DialogModulationParameters


class ModulationParametersDialog(QDialog):
    def __init__(self, parameters: list, modulation_type: str, parent=None):
        super().__init__(parent)
        self.ui = Ui_DialogModulationParameters()
        self.ui.setupUi(self)
        self.setAttribute(Qt.WA_DeleteOnClose)
        self.setWindowFlags(Qt.Window)

        self.parameters = parameters
        self.num_bits = int(math.log2(len(parameters)))

        if "FSK" in modulation_type:
            self.ui.tblSymbolParameters.setItemDelegateForColumn(
                1, KillerSpinBoxDelegate(-1e12, 1e12, self)
            )
            self.ui.tblSymbolParameters.horizontalHeaderItem(1).setText(
                "Frequency in Hz"
            )
        elif "ASK" in modulation_type:
            self.ui.tblSymbolParameters.horizontalHeaderItem(1).setText("Amplitude")
            self.ui.tblSymbolParameters.setItemDelegateForColumn(
                1, SpinBoxDelegate(0, 100, self, "%")
            )
        elif "PSK" in modulation_type:
            self.ui.tblSymbolParameters.setItemDelegateForColumn(
                1, SpinBoxDelegate(-360, 360, self, "°")
            )
            self.ui.tblSymbolParameters.horizontalHeaderItem(1).setText("Phase")

        fmt = "{0:0" + str(self.num_bits) + "b}"
        self.ui.tblSymbolParameters.setRowCount(len(parameters))
        for i, parameter in enumerate(parameters):
            item = QTableWidgetItem(fmt.format(i))
            font = item.font()
            font.setBold(True)
            item.setFont(font)
            item.setFlags(Qt.ItemIsEnabled)
            self.ui.tblSymbolParameters.setItem(i, 0, item)

            item = QTableWidgetItem()
            item.setData(Qt.DisplayRole, self.parameters[i])
            self.ui.tblSymbolParameters.setItem(i, 1, item)
            self.ui.tblSymbolParameters.openPersistentEditor(
                self.ui.tblSymbolParameters.item(i, 1)
            )

        self.create_connects()

    def create_connects(self):
        self.ui.buttonBox.accepted.connect(self.on_accepted)
        self.ui.buttonBox.rejected.connect(self.reject)

    @pyqtSlot()
    def on_accepted(self):
        for i in range(self.ui.tblSymbolParameters.rowCount()):
            self.parameters[i] = float(self.ui.tblSymbolParameters.item(i, 1).text())

        self.accept()


if __name__ == "__main__":
    from PyQt5.QtWidgets import QApplication

    app = QApplication(["urh"])

    dialog = ModulationParametersDialog([0, 100.0], "ASK")
    dialog.show()

    app.exec_()



import csv

import os
import numpy as np
from PyQt5.QtCore import Qt, pyqtSlot, pyqtSignal
from PyQt5.QtWidgets import (
    QDialog,
    QInputDialog,
    QApplication,
    QCompleter,
    QDirModel,
    QFileDialog,
)

from urh.ui.ui_csv_wizard import Ui_DialogCSVImport
from urh.util import FileOperator, util
from urh.util.Errors import Errors


class CSVImportDialog(QDialog):
    data_imported = pyqtSignal(str, float)  # Complex Filename + Sample Rate

    PREVIEW_ROWS = 100
    COLUMNS = {"T": 0, "I": 1, "Q": 2}

    def __init__(self, filename="", parent=None):
        super().__init__(parent)
        self.ui = Ui_DialogCSVImport()
        self.ui.setupUi(self)
        self.setAttribute(Qt.WA_DeleteOnClose)
        self.setWindowFlags(Qt.Window)

        self.ui.btnAutoDefault.hide()

        completer = QCompleter()
        completer.setModel(QDirModel(completer))
        self.ui.lineEditFilename.setCompleter(completer)

        self.filename = None  # type: str
        self.ui.lineEditFilename.setText(filename)
        self.update_file()

        self.ui.tableWidgetPreview.setColumnHidden(self.COLUMNS["T"], True)
        self.update_preview()

        self.create_connects()

    def create_connects(self):
        self.accepted.connect(self.on_accepted)
        self.ui.lineEditFilename.editingFinished.connect(
            self.on_line_edit_filename_editing_finished
        )
        self.ui.btnChooseFile.clicked.connect(self.on_btn_choose_file_clicked)
        self.ui.btnAddSeparator.clicked.connect(self.on_btn_add_separator_clicked)
        self.ui.comboBoxCSVSeparator.currentIndexChanged.connect(
            self.on_combobox_csv_separator_current_index_changed
        )
        self.ui.spinBoxIDataColumn.valueChanged.connect(
            self.on_spinbox_i_data_column_value_changed
        )
        self.ui.spinBoxQDataColumn.valueChanged.connect(
            self.on_spinbox_q_data_column_value_changed
        )
        self.ui.spinBoxTimestampColumn.valueChanged.connect(
            self.on_spinbox_timestamp_value_changed
        )

    def update_file(self):
        filename = self.ui.lineEditFilename.text()
        self.filename = filename

        enable = util.file_can_be_opened(filename)
        if enable:
            with open(self.filename, encoding="utf-8-sig") as f:
                lines = []
                for i, line in enumerate(f):
                    if i >= self.PREVIEW_ROWS:
                        break
                    lines.append(line.strip())
                self.ui.plainTextEditFilePreview.setPlainText("\n".join(lines))
        else:
            self.ui.plainTextEditFilePreview.clear()

        self.ui.plainTextEditFilePreview.setEnabled(enable)
        self.ui.comboBoxCSVSeparator.setEnabled(enable)
        self.ui.spinBoxIDataColumn.setEnabled(enable)
        self.ui.spinBoxQDataColumn.setEnabled(enable)
        self.ui.spinBoxTimestampColumn.setEnabled(enable)
        self.ui.tableWidgetPreview.setEnabled(enable)
        self.ui.labelFileNotFound.setVisible(not enable)

    def update_preview(self):
        if not util.file_can_be_opened(self.filename):
            self.update_file()
            return

        i_data_col = self.ui.spinBoxIDataColumn.value() - 1
        q_data_col = self.ui.spinBoxQDataColumn.value() - 1
        timestamp_col = self.ui.spinBoxTimestampColumn.value() - 1

        self.ui.tableWidgetPreview.setRowCount(self.PREVIEW_ROWS)

        with open(self.filename, encoding="utf-8-sig") as f:
            csv_reader = csv.reader(
                f, delimiter=self.ui.comboBoxCSVSeparator.currentText()
            )
            row = -1

            for line in csv_reader:
                row += 1
                result = self.parse_csv_line(
                    line, i_data_col, q_data_col, timestamp_col
                )
                if result is not None:
                    for key, value in result.items():
                        self.ui.tableWidgetPreview.setItem(
                            row, self.COLUMNS[key], util.create_table_item(value)
                        )
                else:
                    for col in self.COLUMNS.values():
                        self.ui.tableWidgetPreview.setItem(
                            row, col, util.create_table_item("Invalid")
                        )

                if row >= self.PREVIEW_ROWS - 1:
                    break

            self.ui.tableWidgetPreview.setRowCount(row + 1)

    @staticmethod
    def parse_csv_line(
        csv_line: str, i_data_col: int, q_data_col: int, timestamp_col: int
    ):
        result = dict()

        if i_data_col >= 0:
            try:
                result["I"] = float(csv_line[i_data_col])
            except:
                return None
        else:
            result["I"] = 0.0

        if q_data_col >= 0:
            try:
                result["Q"] = float(csv_line[q_data_col])
            except:
                return None
        else:
            result["Q"] = 0.0

        if timestamp_col >= 0:
            try:
                result["T"] = float(csv_line[timestamp_col])
            except:
                return None

        return result

    @staticmethod
    def parse_csv_file(
        filename: str, separator: str, i_data_col: int, q_data_col=-1, t_data_col=-1
    ):
        iq_data = []
        timestamps = [] if t_data_col > -1 else None
        with open(filename, encoding="utf-8-sig") as f:
            csv_reader = csv.reader(f, delimiter=separator)
            for line in csv_reader:
                parsed = CSVImportDialog.parse_csv_line(
                    line, i_data_col, q_data_col, t_data_col
                )
                if parsed is None:
                    continue

                iq_data.append(complex(parsed["I"], parsed["Q"]))
                if timestamps is not None:
                    timestamps.append(parsed["T"])

        iq_data = np.asarray(iq_data, dtype=np.complex64)
        sample_rate = CSVImportDialog.estimate_sample_rate(timestamps)
        return iq_data / abs(iq_data.max()), sample_rate

    @staticmethod
    def estimate_sample_rate(timestamps):
        if timestamps is None or len(timestamps) < 2:
            return None

        previous_timestamp = timestamps[0]
        durations = []

        for timestamp in timestamps[1 : CSVImportDialog.PREVIEW_ROWS]:
            durations.append(abs(timestamp - previous_timestamp))
            previous_timestamp = timestamp

        return 1 / (sum(durations) / len(durations))

    @pyqtSlot()
    def on_line_edit_filename_editing_finished(self):
        self.update_file()
        self.update_preview()

    @pyqtSlot()
    def on_btn_choose_file_clicked(self):
        filename, _ = QFileDialog.getOpenFileName(
            self,
            self.tr("Choose file"),
            directory=FileOperator.RECENT_PATH,
            filter="CSV files (*.csv);;All files (*.*)",
        )

        if filename:
            self.ui.lineEditFilename.setText(filename)
            self.ui.lineEditFilename.editingFinished.emit()

    @pyqtSlot()
    def on_btn_add_separator_clicked(self):
        sep, ok = QInputDialog.getText(self, "Enter Separator", "Separator:", text=",")
        if ok and sep not in (
            self.ui.comboBoxCSVSeparator.itemText(i)
            for i in range(self.ui.comboBoxCSVSeparator.count())
        ):
            if len(sep) == 1:
                self.ui.comboBoxCSVSeparator.addItem(sep)
            else:
                Errors.generic_error(
                    "Invalid Separator", "Separator must be exactly one character."
                )

    @pyqtSlot(int)
    def on_combobox_csv_separator_current_index_changed(self, index: int):
        self.update_preview()

    @pyqtSlot(int)
    def on_spinbox_i_data_column_value_changed(self, value: int):
        self.update_preview()

    @pyqtSlot(int)
    def on_spinbox_q_data_column_value_changed(self, value: int):
        self.update_preview()

    @pyqtSlot(int)
    def on_spinbox_timestamp_value_changed(self, value: int):
        self.ui.tableWidgetPreview.setColumnHidden(self.COLUMNS["T"], value == 0)
        self.update_preview()

    @pyqtSlot()
    def on_accepted(self):
        QApplication.setOverrideCursor(Qt.WaitCursor)

        iq_data, sample_rate = self.parse_csv_file(
            self.filename,
            self.ui.comboBoxCSVSeparator.currentText(),
            self.ui.spinBoxIDataColumn.value() - 1,
            self.ui.spinBoxQDataColumn.value() - 1,
            self.ui.spinBoxTimestampColumn.value() - 1,
        )

        target_filename = self.filename.rstrip(".csv")
        if os.path.exists(target_filename + ".complex"):
            i = 1
            while os.path.exists(target_filename + "_" + str(i) + ".complex"):
                i += 1
        else:
            i = None

        target_filename = target_filename if not i else target_filename + "_" + str(i)
        target_filename += ".complex"

        iq_data.tofile(target_filename)

        self.data_imported.emit(
            target_filename, sample_rate if sample_rate is not None else 0
        )
        QApplication.restoreOverrideCursor()


if __name__ == "__main__":
    app = QApplication(["urh"])
    csv_dia = CSVImportDialog()
    csv_dia.exec_()



import xml.etree.ElementTree as ET

from urh.simulator.SimulatorCounterAction import SimulatorCounterAction
from urh.simulator.SimulatorItem import SimulatorItem
from urh.simulator.SimulatorProtocolLabel import SimulatorProtocolLabel
from urh.simulator.SimulatorRule import (
    SimulatorRule,
    SimulatorRuleCondition,
    ConditionType,
)
from urh.simulator.SimulatorTriggerCommandAction import SimulatorTriggerCommandAction


class SimulatorGotoAction(SimulatorItem):
    def __init__(self):
        super().__init__()
        self.goto_target = None  # type: str

    def set_parent(self, value):
        if value is not None:
            assert value.parent() is None or isinstance(value, SimulatorRuleCondition)

        super().set_parent(value)

    @property
    def target(self):
        return (
            self.simulator_config.item_dict[self.goto_target]
            if self.validate()
            else None
        )

    def validate(self):
        target = self.simulator_config.item_dict.get(self.goto_target, None)
        return self.is_valid_goto_target(self.goto_target, target)

    def get_valid_goto_targets(self):
        valid_targets = []

        for key, value in self.simulator_config.item_dict.items():
            if value != self and SimulatorGotoAction.is_valid_goto_target(key, value):
                valid_targets.append(key)

        return valid_targets

    def to_xml(self) -> ET.Element:
        attributes = dict()
        if self.goto_target is not None:
            attributes["goto_target"] = self.goto_target

        return ET.Element("simulator_goto_action", attrib=attributes)

    @classmethod
    def from_xml(cls, tag: ET.Element):
        result = SimulatorGotoAction()
        result.goto_target = tag.get("goto_target", None)
        return result

    @staticmethod
    def is_valid_goto_target(caption: str, item: SimulatorItem):
        if item is None:
            return False
        if isinstance(item, SimulatorProtocolLabel) or isinstance(item, SimulatorRule):
            return False
        if isinstance(item, SimulatorRuleCondition) and item.type != ConditionType.IF:
            return False
        if isinstance(item, SimulatorCounterAction):
            return False
        if isinstance(item, SimulatorTriggerCommandAction) and caption.endswith("rc"):
            return False

        return True



INDENT_WIDTH_PX = 20


def monospace(string):
    return "<samp>" + string + "</samp>"


def indent_string(string, depth=1):
    width = depth * INDENT_WIDTH_PX
    return '<table style="margin-left: {0}px;" border=0><tr><td>{1}</td></tr></table>'.format(
        width, string
    )


def mark_differences(value: str, compare_against: str):
    result = []
    for i, char in enumerate(value):
        try:
            if char != compare_against[i]:
                result.append('<font color="red">{}</font>'.format(char))
            else:
                result.append(char)
        except IndexError:
            result.append(char)

    return "".join(result)


def align_expected_and_got_value(expected: str, got: str, align_depth=1):
    width = align_depth * INDENT_WIDTH_PX
    got_marked = mark_differences(got, expected)
    return (
        '<table style="margin-left: {0}px;" border=0>'
        "<tr><td>Expected: </td><td>{1}</td></tr><tr><td>Got: </td><td>{2}</td> </tr>"
        "</table>".format(width, monospace(expected), monospace(got_marked))
    )



import math

import numpy as np
from PyQt5.QtGui import QImage

from urh import colormaps
from urh.cythonext import util
from urh.signalprocessing.IQArray import IQArray
from urh.util.Logger import logger


class Spectrogram(object):
    MAX_LINES_PER_VIEW = 1000
    DEFAULT_FFT_WINDOW_SIZE = 1024

    def __init__(
        self,
        samples: np.ndarray,
        window_size=DEFAULT_FFT_WINDOW_SIZE,
        overlap_factor=0.5,
        window_function=np.hanning,
    ):
        """

        :param samples: Complex samples
        :param window_size: Size of DFT window
        :param overlap_factor: Value between 0 (= No Overlapping) and 1 (= Full overlapping) of windows
        :param window_function: Function for DFT window
        """

        self.__samples = np.zeros(1, dtype=np.complex64)
        self.samples = samples

        self.__window_size = window_size
        self.__overlap_factor = overlap_factor
        self.__window_function = window_function

        self.data_min, self.data_max = -140, 10

    @property
    def samples(self):
        return self.__samples

    @samples.setter
    def samples(self, value):
        if isinstance(value, IQArray):
            value = value.as_complex64()
        elif isinstance(value, np.ndarray) and value.dtype != np.complex64:
            value = IQArray(value).as_complex64()
        elif value is None:
            value = np.zeros(1, dtype=np.complex64)

        self.__samples = value

    @property
    def window_size(self):
        return self.__window_size

    @window_size.setter
    def window_size(self, value):
        self.__window_size = value

    @property
    def overlap_factor(self):
        return self.__overlap_factor

    @overlap_factor.setter
    def overlap_factor(self, value):
        self.__overlap_factor = value

    @property
    def window_function(self):
        return self.__window_function

    @window_function.setter
    def window_function(self, value):
        self.__window_function = value

    @property
    def time_bins(self):
        return int(math.ceil(len(self.samples) / self.hop_size))

    @property
    def freq_bins(self):
        return self.window_size

    @property
    def hop_size(self):
        """
        hop size determines by how many samples the window is advanced
        """
        return self.window_size - int(self.overlap_factor * self.window_size)

    def stft(self, samples: np.ndarray):
        """
        Perform Short-time Fourier transform to get the spectrogram for the given samples
        :return: short-time Fourier transform of the given signal
        """
        window = self.window_function(self.window_size)
        hop_size = self.hop_size

        if len(samples) < self.window_size:
            samples = np.append(samples, np.zeros(self.window_size - len(samples)))

        num_frames = max(1, (len(samples) - self.window_size) // hop_size + 1)

        # Get frames as numpy view with stride_tricks to save RAM
        # Same as: frames = [padded_samples[i*hop_size:i*hop_size+self.window_size] for i in range(num_frames)]
        shape = (num_frames, self.window_size)
        strides = (hop_size * samples.strides[-1], samples.strides[-1])
        frames = np.lib.stride_tricks.as_strided(samples, shape=shape, strides=strides)

        result = np.fft.fft(frames * window, self.window_size) / np.atleast_1d(
            self.window_size
        )
        return result

    def export_to_fta(self, sample_rate, filename: str, include_amplitude=False):
        """
        Export to Frequency, Time, Amplitude file.
        Frequency is double, Time (nanosecond) is uint32, Amplitude is float32

        :return:
        """
        spectrogram = self.__calculate_spectrogram(self.samples)
        spectrogram = np.flipud(spectrogram.T)
        if include_amplitude:
            result = np.empty(
                (spectrogram.shape[0], spectrogram.shape[1], 3),
                dtype=[("f", np.float64), ("t", np.uint32), ("a", np.float32)],
            )
        else:
            result = np.empty(
                (spectrogram.shape[0], spectrogram.shape[1], 2),
                dtype=[("f", np.float64), ("t", np.uint32)],
            )

        fft_freqs = np.fft.fftshift(
            np.fft.fftfreq(spectrogram.shape[0], 1 / sample_rate)
        )
        time_width = 1e9 * ((len(self.samples) / sample_rate) / spectrogram.shape[1])

        for i in range(spectrogram.shape[0]):
            for j in range(spectrogram.shape[1]):
                if include_amplitude:
                    result[i, j] = (
                        fft_freqs[i],
                        int(j * time_width),
                        spectrogram[i, j],
                    )
                else:
                    result[i, j] = (fft_freqs[i], int(j * time_width))

        result.tofile(filename)

    def __calculate_spectrogram(self, samples: np.ndarray) -> np.ndarray:
        # Only shift axis 1 (frequency) and not time
        spectrogram = np.fft.fftshift(self.stft(samples), axes=(1,))
        spectrogram = util.arr2decibel(spectrogram.astype(np.complex64))

        # Flip Array so Y axis goes from negative to positive
        return np.fliplr(spectrogram)

    def create_spectrogram_image(
        self,
        sample_start: int = None,
        sample_end: int = None,
        step: int = None,
        transpose=False,
    ):
        spectrogram = self.__calculate_spectrogram(
            self.samples[sample_start:sample_end:step]
        )
        if transpose:
            spectrogram = np.flipud(spectrogram.T)
        return self.create_image(
            spectrogram,
            colormaps.chosen_colormap_numpy_bgra,
            self.data_min,
            self.data_max,
        )

    def create_image_segments(self):
        n_segments = max(1, self.time_bins // self.MAX_LINES_PER_VIEW)
        step = self.time_bins / n_segments
        step = max(1, int((step / self.hop_size) * self.hop_size**2))

        for i in range(0, len(self.samples), step):
            image = self.create_spectrogram_image(sample_start=i, sample_end=i + step)
            yield image

    @staticmethod
    def apply_bgra_lookup(
        data: np.ndarray, colormap, data_min=None, data_max=None, normalize=True
    ) -> np.ndarray:
        if normalize and (data_min is None or data_max is None):
            raise ValueError("Can't normalize without data min and data max")

        if normalize:
            normalized_values = (len(colormap) - 1) * (
                (data.T - data_min) / (data_max - data_min)
            )
        else:
            normalized_values = data.T

        return np.take(colormap, normalized_values.astype(int), axis=0, mode="clip")

    @staticmethod
    def create_image(
        data: np.ndarray, colormap, data_min=None, data_max=None, normalize=True
    ) -> QImage:
        """
        Create QImage from ARGB array.
        The ARGB must have shape (width, height, 4) and dtype=ubyte.
        NOTE: The order of values in the 3rd axis must be (blue, green, red, alpha).
        :return:
        """
        image_data = Spectrogram.apply_bgra_lookup(
            data, colormap, data_min, data_max, normalize
        )

        if not image_data.flags["C_CONTIGUOUS"]:
            logger.debug("Array was not C_CONTIGUOUS. Converting it.")
            image_data = np.ascontiguousarray(image_data)

        try:
            # QImage constructor needs inverted row/column order
            image = QImage(
                image_data.ctypes.data,
                image_data.shape[1],
                image_data.shape[0],
                QImage.Format_ARGB32,
            )
        except Exception as e:
            logger.error("could not create image " + str(e))
            return QImage()

        image.data = image_data
        return image

    @staticmethod
    def create_colormap_image(colormap_name: str, height=100) -> QImage:
        colormap = colormaps.calculate_numpy_brga_for(colormap_name)

        indices = np.zeros((len(colormap), height), dtype=np.int64)
        for i in np.arange(len(colormap), dtype=np.int64):
            indices[i, :] = np.repeat(i, height)

        return Spectrogram.create_image(indices, colormap, normalize=False)



from PyQt5.QtCore import pyqtSignal
from PyQt5.QtGui import QDropEvent, QDragEnterEvent, QWheelEvent
from PyQt5.QtWidgets import QScrollArea


class ScrollArea(QScrollArea):
    files_dropped = pyqtSignal(list)

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setAcceptDrops(True)

    def dropEvent(self, event: QDropEvent):
        self.files_dropped.emit(event.mimeData().urls())

    def dragEnterEvent(self, event: QDragEnterEvent):
        event.accept()

    def wheelEvent(self, event: QWheelEvent):
        event.ignore()



from PyQt5.QtGui import QMouseEvent
from PyQt5.QtCore import Qt, pyqtSlot, QModelIndex, pyqtSignal

from urh.models.SimulatorMessageFieldModel import SimulatorMessageFieldModel
from urh.ui.views.ProtocolLabelTableView import ProtocolLabelTableView


class SimulatorLabelTableView(ProtocolLabelTableView):
    item_link_clicked = pyqtSignal(int, int)

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setMouseTracking(True)
        self.clicked.connect(self.on_clicked)

    def model(self) -> SimulatorMessageFieldModel:
        return super().model()

    def mouseMoveEvent(self, e: QMouseEvent):
        index = self.indexAt(e.pos())
        if self.model().link_index(index):
            self.setCursor(Qt.PointingHandCursor)
        else:
            self.unsetCursor()

    @pyqtSlot(QModelIndex)
    def on_clicked(self, index: QModelIndex):
        if self.model().link_index(index):
            self.item_link_clicked.emit(index.row(), index.column())



import copy
import math
from collections import defaultdict

import numpy as np

from urh.awre import AutoAssigner
from urh.awre.CommonRange import (
    CommonRange,
    EmptyCommonRange,
    CommonRangeContainer,
    ChecksumRange,
)
from urh.awre.Preprocessor import Preprocessor
from urh.awre.engines.AddressEngine import AddressEngine
from urh.awre.engines.ChecksumEngine import ChecksumEngine
from urh.awre.engines.LengthEngine import LengthEngine
from urh.awre.engines.SequenceNumberEngine import SequenceNumberEngine
from urh.cythonext import awre_util
from urh.signalprocessing.ChecksumLabel import ChecksumLabel
from urh.signalprocessing.FieldType import FieldType
from urh.signalprocessing.Message import Message
from urh.signalprocessing.MessageType import MessageType
from urh.signalprocessing.ProtocoLabel import ProtocolLabel
from urh.util.WSPChecksum import WSPChecksum


class FormatFinder(object):
    MIN_MESSAGES_PER_CLUSTER = 2

    def __init__(self, messages, participants=None, shortest_field_length=None):
        """

        :type messages: list of Message
        :param participants:
        """
        if participants is not None:
            AutoAssigner.auto_assign_participants(messages, participants)

        existing_message_types_by_msg = {
            i: msg.message_type for i, msg in enumerate(messages)
        }
        self.existing_message_types = defaultdict(list)
        for i, message_type in existing_message_types_by_msg.items():
            self.existing_message_types[message_type].append(i)

        preprocessor = Preprocessor(
            self.get_bitvectors_from_messages(messages), existing_message_types_by_msg
        )
        (
            self.preamble_starts,
            self.preamble_lengths,
            sync_len,
        ) = preprocessor.preprocess()
        self.sync_ends = self.preamble_starts + self.preamble_lengths + sync_len

        n = shortest_field_length
        if n is None:
            # 0 = no sync found
            n = (
                8
                if sync_len >= 8
                else 4
                if sync_len >= 4
                else 1
                if sync_len >= 1
                else 0
            )

        for i, value in enumerate(self.sync_ends):
            # In doubt it is better to under estimate the sync end
            if n > 0:
                self.sync_ends[i] = (
                    n * max(int(math.floor((value - self.preamble_starts[i]) / n)), 1)
                    + self.preamble_starts[i]
                )
            else:
                self.sync_ends[i] = self.preamble_starts[i]

            if self.sync_ends[i] - self.preamble_starts[i] < self.preamble_lengths[i]:
                self.preamble_lengths[i] = self.sync_ends[i] - self.preamble_starts[i]

        self.bitvectors = self.get_bitvectors_from_messages(messages, self.sync_ends)
        self.hexvectors = self.get_hexvectors(self.bitvectors)
        self.current_iteration = 0

        participants = list(
            sorted(
                set(msg.participant for msg in messages if msg.participant is not None)
            )
        )
        self.participant_indices = [
            participants.index(msg.participant) if msg.participant is not None else -1
            for msg in messages
        ]
        self.known_participant_addresses = {
            participants.index(p): np.array(
                [int(h, 16) for h in p.address_hex], dtype=np.uint8
            )
            for p in participants
            if p and p.address_hex
        }

    @property
    def message_types(self):
        """

        :rtype: list of MessageType
        """
        return sorted(self.existing_message_types.keys(), key=lambda x: x.name)

    def perform_iteration_for_message_type(self, message_type: MessageType):
        """
        Perform a field inference iteration for messages of the given message type
        This routine will return newly found fields as a set of Common Ranges

        :param message_type:
        :rtype: set of CommonRange
        """
        indices = self.existing_message_types[message_type]
        engines = []

        # We can take an arbitrary sync end to correct the already labeled fields for this message type,
        # because if the existing labels would have different sync positions,
        # they would not belong to the same message type in the first place
        sync_end = self.sync_ends[indices[0]] if indices else 0
        already_labeled = [
            (lbl.start - sync_end, lbl.end - sync_end)
            for lbl in message_type
            if lbl.start >= sync_end
        ]

        if not message_type.get_first_label_with_type(FieldType.Function.LENGTH):
            engines.append(
                LengthEngine(
                    [self.bitvectors[i] for i in indices],
                    already_labeled=already_labeled,
                )
            )

        if not message_type.get_first_label_with_type(FieldType.Function.SRC_ADDRESS):
            engines.append(
                AddressEngine(
                    [self.hexvectors[i] for i in indices],
                    [self.participant_indices[i] for i in indices],
                    self.known_participant_addresses,
                    already_labeled=already_labeled,
                )
            )
        elif not message_type.get_first_label_with_type(FieldType.Function.DST_ADDRESS):
            engines.append(
                AddressEngine(
                    [self.hexvectors[i] for i in indices],
                    [self.participant_indices[i] for i in indices],
                    self.known_participant_addresses,
                    already_labeled=already_labeled,
                    src_field_present=True,
                )
            )

        if not message_type.get_first_label_with_type(
            FieldType.Function.SEQUENCE_NUMBER
        ):
            engines.append(
                SequenceNumberEngine(
                    [self.bitvectors[i] for i in indices],
                    already_labeled=already_labeled,
                )
            )
        if not message_type.get_first_label_with_type(FieldType.Function.CHECKSUM):
            # If checksum was not found in first iteration, it will also not be found in next one
            if self.current_iteration == 0:
                engines.append(
                    ChecksumEngine(
                        [self.bitvectors[i] for i in indices],
                        already_labeled=already_labeled,
                    )
                )

        result = set()
        for engine in engines:
            high_scored_ranges = engine.find()  # type: list[CommonRange]
            high_scored_ranges = self.retransform_message_indices(
                high_scored_ranges, indices, self.sync_ends
            )
            merged_ranges = self.merge_common_ranges(high_scored_ranges)
            result.update(merged_ranges)
        return result

    def perform_iteration(self) -> bool:
        new_field_found = False

        for message_type in self.existing_message_types.copy():
            new_fields_for_message_type = self.perform_iteration_for_message_type(
                message_type
            )
            new_fields_for_message_type.update(
                self.get_preamble_and_sync(
                    self.preamble_starts,
                    self.preamble_lengths,
                    self.sync_ends,
                    message_type_indices=self.existing_message_types[message_type],
                )
            )

            self.remove_overlapping_fields(new_fields_for_message_type, message_type)
            containers = self.create_common_range_containers(
                new_fields_for_message_type
            )

            # Store addresses of participants if we found a SRC address field
            participants_with_unknown_address = set(self.participant_indices) - set(
                self.known_participant_addresses
            )
            participants_with_unknown_address.discard(-1)

            if participants_with_unknown_address:
                for container in containers:
                    src_range = next(
                        (
                            rng
                            for rng in container
                            if rng.field_type == "source address"
                        ),
                        None,
                    )
                    if src_range is None:
                        continue
                    for msg_index in src_range.message_indices:
                        if len(participants_with_unknown_address) == 0:
                            break
                        p = self.participant_indices[msg_index]
                        if p not in self.known_participant_addresses:
                            hex_vector = self.hexvectors[msg_index]
                            self.known_participant_addresses[p] = hex_vector[
                                src_range.start : src_range.end + 1
                            ]
                            participants_with_unknown_address.discard(p)

            new_field_found |= len(containers) > 0

            if len(containers) == 1:
                for rng in containers[0]:
                    self.add_range_to_message_type(rng, message_type)
            elif len(containers) > 1:
                del self.existing_message_types[message_type]

                for i, container in enumerate(containers):
                    new_message_type = copy.deepcopy(message_type)  # type: MessageType

                    if i > 0:
                        new_message_type.name = "Message Type {}.{}".format(
                            self.current_iteration + 1, i
                        )
                        new_message_type.give_new_id()

                    for rng in container:
                        self.add_range_to_message_type(rng, new_message_type)

                    self.existing_message_types[new_message_type].extend(
                        sorted(container.message_indices)
                    )

        return new_field_found

    def run(self, max_iterations=10):
        self.current_iteration = 0
        while self.perform_iteration() and self.current_iteration < max_iterations:
            self.current_iteration += 1

        if len(self.message_types) > 0:
            messages_without_message_type = set(range(len(self.bitvectors))) - set(
                i for l in self.existing_message_types.values() for i in l
            )

            # add to default message type
            self.existing_message_types[self.message_types[0]].extend(
                list(messages_without_message_type)
            )

    @staticmethod
    def remove_overlapping_fields(common_ranges, message_type: MessageType):
        """
        Remove all fields from a set of CommonRanges which overlap with fields of the existing message type

        :type common_ranges: set of CommonRange
        :param message_type:
        :return:
        """
        if len(message_type) == 0:
            return

        for rng in common_ranges.copy():
            for lbl in message_type:  # type: ProtocolLabel
                if any(
                    i in range(rng.bit_start, rng.bit_end)
                    for i in range(lbl.start, lbl.end)
                ):
                    common_ranges.discard(rng)
                    break

    @staticmethod
    def merge_common_ranges(common_ranges):
        """
        Merge common ranges if possible

        :type common_ranges: list of CommonRange
        :rtype: list of CommonRange
        """
        merged_ranges = []
        for common_range in common_ranges:
            assert isinstance(common_range, CommonRange)
            try:
                same_range = next(
                    rng
                    for rng in merged_ranges
                    if rng.bit_start == common_range.bit_start
                    and rng.bit_end == common_range.bit_end
                    and rng.field_type == common_range.field_type
                )
                same_range.values.extend(common_range.values)
                same_range.message_indices.update(common_range.message_indices)
            except StopIteration:
                merged_ranges.append(common_range)

        return merged_ranges

    @staticmethod
    def add_range_to_message_type(common_range: CommonRange, message_type: MessageType):
        field_type = FieldType.from_caption(common_range.field_type)
        label = message_type.add_protocol_label(
            name=common_range.field_type,
            start=common_range.bit_start,
            end=common_range.bit_end,
            auto_created=True,
            type=field_type,
        )
        label.display_endianness = common_range.byte_order

        if field_type.function == FieldType.Function.CHECKSUM:
            assert isinstance(label, ChecksumLabel)
            assert isinstance(common_range, ChecksumRange)
            label.data_ranges = [
                [common_range.data_range_bit_start, common_range.data_range_bit_end]
            ]

            if isinstance(common_range.crc, WSPChecksum):
                label.category = ChecksumLabel.Category.wsp
            else:
                label.checksum = copy.copy(common_range.crc)

    @staticmethod
    def get_hexvectors(bitvectors: list):
        result = awre_util.get_hexvectors(bitvectors)
        return result

    @staticmethod
    def get_bitvectors_from_messages(messages: list, sync_ends: np.ndarray = None):
        if sync_ends is None:
            sync_ends = defaultdict(lambda: None)

        return [
            np.array(msg.decoded_bits[sync_ends[i] :], dtype=np.uint8, order="C")
            for i, msg in enumerate(messages)
        ]

    @staticmethod
    def create_common_range_containers(label_set: set, num_messages: int = None):
        """
        Create message types from set of labels.
        Handle overlapping conflicts and create multiple message types if needed

        :param label_set:
        :param num_messages:
        :return:
        :rtype: list of CommonRangeContainer
        """
        if num_messages is None:
            message_indices = sorted(
                set(i for rng in label_set for i in rng.message_indices)
            )
        else:
            message_indices = range(num_messages)

        result = []
        for i in message_indices:
            labels = sorted(
                set(
                    rng
                    for rng in label_set
                    if i in rng.message_indices
                    and not isinstance(rng, EmptyCommonRange)
                )
            )

            container = next(
                (
                    container
                    for container in result
                    if container.has_same_ranges(labels)
                ),
                None,
            )
            if container is None:
                result.append(CommonRangeContainer(labels, message_indices={i}))
            else:
                container.message_indices.add(i)

        result = FormatFinder.handle_overlapping_conflict(result)

        return result

    @staticmethod
    def handle_overlapping_conflict(containers):
        """
        Handle overlapping conflicts for a list of CommonRangeContainers

        :type containers: list of CommonRangeContainer
        :return:
        """
        result = []
        for container in containers:
            if container.ranges_overlap:
                conflicted_handled = (
                    FormatFinder.__handle_container_overlapping_conflict(container)
                )
            else:
                conflicted_handled = container

            try:
                same_rng_container = next(
                    c
                    for c in result
                    if c.has_same_ranges_as_container(conflicted_handled)
                )
                same_rng_container.message_indices.update(
                    conflicted_handled.message_indices
                )
            except StopIteration:
                result.append(conflicted_handled)

        return result

    @staticmethod
    def __handle_container_overlapping_conflict(container: CommonRangeContainer):
        """
        Handle overlapping conflict for a CommRangeContainer.
        We can assert that all labels in the container share the same message indices
        because we partitioned them in a step before.
        If two or more labels overlap we have three ways to resolve the conflict:

        1. Choose the range with the highest score
        2. If multiple ranges overlap choose the ranges that maximize the overall (cumulated) score
        3. If the overlapping is very small i.e. only 1 or 2 bits we can adjust the start/end of the conflicting ranges

        The ranges inside the container _must_ be sorted i.e. the range with lowest start must be at front

        :param container:
        :return:
        """
        partitions = []  # type: list[list[CommonRange]]
        # partition the container into overlapping partitions
        # results in something like [[A], [B,C], [D], [E,F,G]]] where B and C and E, F, G are overlapping
        for cur_rng in container:
            if len(partitions) == 0:
                partitions.append([cur_rng])
                continue

            last_rng = partitions[-1][-1]  # type: CommonRange
            if cur_rng.overlaps_with(last_rng):
                partitions[-1].append(cur_rng)
            else:
                partitions.append([cur_rng])

        # Todo: Adjust start/end of conflicting ranges if overlapping is very small (i.e. 1 or 2 bits)

        result = []
        # Go through these partitions and handle overlapping conflicts
        for partition in partitions:
            possible_solutions = []
            for i, rng in enumerate(partition):
                # Append every range to this solution that does not overlap with current rng
                solution = [rng] + [
                    r for r in partition[i + 1 :] if not rng.overlaps_with(r)
                ]
                possible_solutions.append(solution)

            # Take solution that maximizes score. In case of tie, choose solution with shorter total length.
            # if there is still a tie prefer solution that contains a length field as is is very likely to be correct
            # if nothing else helps break tie by names of field types to prevent randomness
            best_solution = max(
                possible_solutions,
                key=lambda sol: (
                    sum(r.score for r in sol),
                    -sum(r.length_in_bits for r in sol),
                    "length" in {r.field_type for r in sol},
                    "".join(r.field_type[0] for r in sol),
                ),
            )
            result.extend(best_solution)

        return CommonRangeContainer(result, message_indices=container.message_indices)

    @staticmethod
    def retransform_message_indices(
        common_ranges, message_type_indices: list, sync_ends
    ) -> list:
        """
        Retransform the found message indices of an engine to the original index space
        based on the message indices of the message type.

        Furthermore, set the sync_end of the common ranges so bit_start and bit_end
        match the position in the original space

        :type common_ranges: list of CommonRange
        :param message_type_indices: Messages belonging to the message type the engine ran for
        :type sync_ends: np.ndarray
        :return:
        """
        result = []
        for common_range in common_ranges:
            # Retransform message indices into original space
            message_indices = np.fromiter(
                (message_type_indices[i] for i in common_range.message_indices),
                dtype=int,
                count=len(common_range.message_indices),
            )

            # If we have different sync_ends we need to create a new common range for each different sync_length
            matching_sync_ends = sync_ends[message_indices]
            for sync_end in np.unique(matching_sync_ends):
                rng = copy.deepcopy(common_range)
                rng.sync_end = sync_end
                rng.message_indices = set(
                    message_indices[np.nonzero(matching_sync_ends == sync_end)]
                )
                result.append(rng)

        return result

    @staticmethod
    def get_preamble_and_sync(
        preamble_starts, preamble_lengths, sync_ends, message_type_indices
    ):
        """
        Get preamble and sync common ranges based on the data

        :type preamble_starts: np.ndarray
        :type preamble_lengths: np.ndarray
        :type sync_ends: np.ndarray
        :type message_type_indices: list
        :rtype: set of CommonRange
        """
        assert len(preamble_starts) == len(preamble_lengths) == len(sync_ends)

        result = set()  # type: set[CommonRange]
        for i in message_type_indices:
            preamble = CommonRange(
                preamble_starts[i],
                preamble_lengths[i],
                field_type="preamble",
                message_indices={i},
            )
            existing_preamble = next((rng for rng in result if preamble == rng), None)
            if existing_preamble is not None:
                existing_preamble.message_indices.add(i)
            elif preamble_lengths[i] > 0:
                result.add(preamble)

            preamble_end = preamble_starts[i] + preamble_lengths[i]
            sync_end = sync_ends[i]
            sync = CommonRange(
                preamble_end,
                sync_end - preamble_end,
                field_type="synchronization",
                message_indices={i},
            )
            existing_sync = next((rng for rng in result if sync == rng), None)
            if existing_sync is not None:
                existing_sync.message_indices.add(i)
            elif sync_end - preamble_end > 0:
                result.add(sync)

        return result



import os
import socket
import sys
import tempfile
import time
from queue import Queue, Empty
from subprocess import Popen, PIPE
from threading import Thread

from PyQt5.QtCore import QThread, pyqtSignal

from urh import settings
from urh.util.Logger import logger

ON_POSIX = "posix" in sys.builtin_module_names


class AbstractBaseThread(QThread):
    started = pyqtSignal()
    stopped = pyqtSignal()
    sender_needs_restart = pyqtSignal()

    def __init__(
        self,
        frequency,
        sample_rate,
        bandwidth,
        gain,
        if_gain,
        baseband_gain,
        receiving: bool,
        ip="127.0.0.1",
        parent=None,
    ):
        super().__init__(parent)
        self.ip = ip
        self.gr_port = 1337
        self._sample_rate = sample_rate
        self._frequency = frequency
        self._gain = gain
        self._if_gain = if_gain
        self._baseband_gain = baseband_gain
        self._bandwidth = bandwidth
        self._freq_correction = 1
        self._direct_sampling_mode = 0
        self._antenna_index = 0
        self._channel_index = 0
        self._receiving = receiving  # False for Sender-Thread
        self.device = "USRP"
        self.current_index = 0

        self.is_in_spectrum_mode = False

        self.socket = None

        self.gr_python_interpreter = settings.read("gr_python_interpreter", "")

        self.queue = Queue()
        self.data = None  # Placeholder for SenderThread
        self.current_iteration = 0  # Counts number of Sendings in SenderThread

        self.gr_process = None

    @property
    def sample_rate(self):
        return self._sample_rate

    @sample_rate.setter
    def sample_rate(self, value):
        self._sample_rate = value
        if self.gr_process:
            try:
                self.gr_process.stdin.write(b"SR:" + bytes(str(value), "utf8") + b"\n")
                self.gr_process.stdin.flush()
            except BrokenPipeError:
                pass

    @property
    def frequency(self):
        return self._frequency

    @frequency.setter
    def frequency(self, value):
        self._frequency = value
        if self.gr_process:
            try:
                self.gr_process.stdin.write(b"F:" + bytes(str(value), "utf8") + b"\n")
                self.gr_process.stdin.flush()
            except BrokenPipeError:
                pass

    @property
    def gain(self):
        return self._gain

    @gain.setter
    def gain(self, value):
        self._gain = value
        if self.gr_process:
            try:
                self.gr_process.stdin.write(b"G:" + bytes(str(value), "utf8") + b"\n")
                self.gr_process.stdin.flush()
            except BrokenPipeError:
                pass

    @property
    def if_gain(self):
        return self._if_gain

    @if_gain.setter
    def if_gain(self, value):
        self._if_gain = value
        if self.gr_process:
            try:
                self.gr_process.stdin.write(b"IFG:" + bytes(str(value), "utf8") + b"\n")
                self.gr_process.stdin.flush()
            except BrokenPipeError:
                pass

    @property
    def baseband_gain(self):
        return self._baseband_gain

    @baseband_gain.setter
    def baseband_gain(self, value):
        self._baseband_gain = value
        if self.gr_process:
            try:
                self.gr_process.stdin.write(b"BBG:" + bytes(str(value), "utf8") + b"\n")
                self.gr_process.stdin.flush()
            except BrokenPipeError:
                pass

    @property
    def bandwidth(self):
        return self._bandwidth

    @bandwidth.setter
    def bandwidth(self, value):
        self._bandwidth = value
        if self.gr_process:
            try:
                self.gr_process.stdin.write(b"BW:" + bytes(str(value), "utf8") + b"\n")
                self.gr_process.stdin.flush()
            except BrokenPipeError:
                pass

    @property
    def freq_correction(self):
        return self._freq_correction

    @freq_correction.setter
    def freq_correction(self, value):
        self._freq_correction = value
        if self.gr_process:
            try:
                self.gr_process.stdin.write(b"FC:" + bytes(str(value), "utf8") + b"\n")
                self.gr_process.stdin.flush()
            except BrokenPipeError:
                pass

    @property
    def channel_index(self):
        return self._channel_index

    @channel_index.setter
    def channel_index(self, value):
        self._channel_index = value

    @property
    def antenna_index(self):
        return self._antenna_index

    @antenna_index.setter
    def antenna_index(self, value):
        self._antenna_index = value

    @property
    def direct_sampling_mode(self):
        return self._direct_sampling_mode

    @direct_sampling_mode.setter
    def direct_sampling_mode(self, value):
        self._direct_sampling_mode = value
        if self.gr_process:
            try:
                self.gr_process.stdin.write(b"DSM:" + bytes(str(value), "utf8") + b"\n")
                self.gr_process.stdin.flush()
            except BrokenPipeError:
                pass

    def initialize_process(self):
        self.started.emit()

        if not hasattr(sys, "frozen"):
            rp = os.path.realpath(os.path.join(os.path.dirname(__file__), "scripts"))
        else:
            rp = os.path.realpath(os.path.dirname(sys.executable))

        suffix = "_recv.py" if self._receiving else "_send.py"
        filename = self.device.lower().split(" ")[0] + suffix

        if not self.gr_python_interpreter:
            self.stop(
                "FATAL: Could not find a GR compatible Python interpreter. "
                "Make sure you have a running GNU Radio installation."
            )
            return

        options = [
            self.gr_python_interpreter,
            os.path.join(rp, filename),
            "--sample-rate",
            str(int(self.sample_rate)),
            "--frequency",
            str(int(self.frequency)),
            "--gain",
            str(self.gain),
            "--if-gain",
            str(self.if_gain),
            "--bb-gain",
            str(self.baseband_gain),
            "--bandwidth",
            str(int(self.bandwidth)),
            "--freq-correction",
            str(self.freq_correction),
            "--direct-sampling",
            str(self.direct_sampling_mode),
            "--channel-index",
            str(self.channel_index),
            "--port",
            str(self.gr_port),
        ]

        logger.info("Starting GNU Radio")
        logger.debug(" ".join(options))
        self.gr_process = Popen(
            options, stdout=PIPE, stderr=PIPE, stdin=PIPE, bufsize=1
        )
        logger.info("Started GNU Radio")
        t = Thread(
            target=self.enqueue_output, args=(self.gr_process.stderr, self.queue)
        )
        t.daemon = True  # thread dies with the program
        t.start()

    def init_recv_socket(self):
        logger.info("Initializing receive socket")
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        logger.info("Initialized receive socket")

        while not self.isInterruptionRequested():
            try:
                time.sleep(0.1)
                logger.info("Trying to get a connection to GNU Radio...")
                self.socket.connect((self.ip, self.gr_port))
                logger.info("Got connection")
                break
            except (ConnectionRefusedError, ConnectionResetError):
                continue
            except Exception as e:
                logger.error("Unexpected error", str(e))

    def run(self):
        pass

    def read_errors(self, initial_errors=None):
        result = [] if initial_errors is None else initial_errors
        while True:
            try:
                result.append(self.queue.get_nowait())
            except Empty:
                break

        result = b"".join(result)
        try:
            return result.decode("utf-8")
        except UnicodeDecodeError:
            return "Could not decode device message"

    def enqueue_output(self, out, queue):
        for line in iter(out.readline, b""):
            queue.put(line)
        out.close()

    def stop(self, msg: str):
        if msg and not msg.startswith("FIN"):
            self.requestInterruption()
            time.sleep(0.1)

        try:
            logger.info("Kill grc process")
            self.gr_process.kill()
            logger.info("Term grc process")
            self.gr_process.terminate()
            self.gr_process = None
        except AttributeError:
            pass

        logger.info(msg)
        self.stopped.emit()



from collections import OrderedDict
from multiprocessing import Array
from multiprocessing.connection import Connection

import numpy as np
import pyaudio

from urh.dev.native.Device import Device
from urh.util.Logger import logger


class SoundCard(Device):
    DEVICE_LIB = pyaudio
    ASYNCHRONOUS = False
    DEVICE_METHODS = dict()

    CHUNK_SIZE = 1024
    SYNC_TX_CHUNK_SIZE = 2 * CHUNK_SIZE
    CONTINUOUS_TX_CHUNK_SIZE = SYNC_TX_CHUNK_SIZE

    SAMPLE_RATE = 48000

    pyaudio_handle = None
    pyaudio_stream = None

    DATA_TYPE = np.float32

    @classmethod
    def init_device(
        cls, ctrl_connection: Connection, is_tx: bool, parameters: OrderedDict
    ) -> bool:
        try:
            cls.SAMPLE_RATE = int(parameters[cls.Command.SET_SAMPLE_RATE.name])
        except (KeyError, ValueError):
            pass
        return super().init_device(ctrl_connection, is_tx, parameters)

    @classmethod
    def setup_device(cls, ctrl_connection: Connection, device_identifier):
        ctrl_connection.send("Initializing pyaudio...")
        try:
            cls.pyaudio_handle = pyaudio.PyAudio()
            ctrl_connection.send("Initialized pyaudio")
            return True
        except Exception as e:
            logger.exception(e)
            ctrl_connection.send("Failed to initialize pyaudio")

    @classmethod
    def prepare_sync_receive(cls, ctrl_connection: Connection):
        try:
            cls.pyaudio_stream = cls.pyaudio_handle.open(
                format=pyaudio.paFloat32,
                channels=2,
                rate=cls.SAMPLE_RATE,
                input=True,
                frames_per_buffer=cls.CHUNK_SIZE,
            )
            ctrl_connection.send("Successfully started pyaudio stream")
            return 0
        except Exception as e:
            logger.exception(e)
            ctrl_connection.send("Failed to start pyaudio stream")

    @classmethod
    def prepare_sync_send(cls, ctrl_connection: Connection):
        try:
            cls.pyaudio_stream = cls.pyaudio_handle.open(
                format=pyaudio.paFloat32,
                channels=2,
                rate=cls.SAMPLE_RATE,
                frames_per_buffer=cls.CHUNK_SIZE,
                output=True,
            )
            ctrl_connection.send("Successfully started pyaudio stream")
            return 0
        except Exception as e:
            logger.exception(e)
            ctrl_connection.send("Failed to start pyaudio stream")

    @classmethod
    def receive_sync(cls, data_conn: Connection):
        if cls.pyaudio_stream:
            data_conn.send_bytes(
                cls.pyaudio_stream.read(cls.CHUNK_SIZE, exception_on_overflow=False)
            )

    @classmethod
    def send_sync(cls, data):
        if cls.pyaudio_stream:
            data_bytes = (
                data.tostring() if isinstance(data, np.ndarray) else bytes(data)
            )
            # pad with zeros if smaller than chunk size
            cls.pyaudio_stream.write(data_bytes.ljust(cls.CHUNK_SIZE * 8, b"\0"))

    @classmethod
    def shutdown_device(cls, ctrl_connection, is_tx: bool):
        logger.debug("shutting down pyaudio...")
        try:
            if cls.pyaudio_stream:
                cls.pyaudio_stream.stop_stream()
                cls.pyaudio_stream.close()
            if cls.pyaudio_handle:
                cls.pyaudio_handle.terminate()
            ctrl_connection.send("CLOSE:0")
        except Exception as e:
            logger.exception(e)
            ctrl_connection.send("Failed to shut down pyaudio")

    def __init__(self, sample_rate, resume_on_full_receive_buffer=False):
        super().__init__(
            center_freq=0,
            sample_rate=sample_rate,
            bandwidth=0,
            gain=1,
            if_gain=1,
            baseband_gain=1,
            resume_on_full_receive_buffer=resume_on_full_receive_buffer,
        )

        self.success = 0
        self.bandwidth_is_adjustable = False

    @property
    def device_parameters(self) -> OrderedDict:
        return OrderedDict(
            [
                (self.Command.SET_SAMPLE_RATE.name, self.sample_rate),
                ("identifier", None),
            ]
        )

    @staticmethod
    def bytes_to_iq(buffer):
        return np.frombuffer(buffer, dtype=np.float32).reshape((-1, 2), order="C")

    @staticmethod
    def iq_to_bytes(samples: np.ndarray):
        arr = Array("f", 2 * len(samples), lock=False)
        numpy_view = np.frombuffer(arr, dtype=np.float32)
        numpy_view[:] = samples.flatten(order="C")
        return arr



import array
import locale
import math
import xml.etree.ElementTree as ET

import numpy as np
from PyQt5.QtGui import QPen
from PyQt5.QtWidgets import QGraphicsScene

from urh import settings
from urh.cythonext import path_creator, signal_functions
from urh.signalprocessing.IQArray import IQArray
from urh.ui.painting.ZoomableScene import ZoomableScene
from urh.util.Formatter import Formatter


class Modulator(object):
    FORCE_DTYPE = None

    MODULATION_TYPES = ["ASK", "FSK", "PSK", "GFSK", "OQPSK"]
    MODULATION_TYPES_VERBOSE = {
        "ASK": "Amplitude Shift Keying (ASK)",
        "FSK": "Frequency Shift Keying (FSK)",
        "PSK": "Phase Shift Keying (PSK)",
        "OQPSK": "Offset Quadrature Phase Shift Keying (OQPSK)",
        "GFSK": "Gaussian Frequeny Shift Keying (GFSK)",
    }

    def __init__(self, name: str):
        self.carrier_freq_hz = 40 * 10**3
        self.carrier_amplitude = 1
        self.carrier_phase_deg = 0
        self.data = [True, False, True, False]
        self.samples_per_symbol = 100
        self.default_sample_rate = 10**6
        self.__sample_rate = None
        self.__modulation_type = "ASK"

        self.__bits_per_symbol = 1

        self.name = name

        self.gauss_bt = 0.5  # bt product for gaussian filter (GFSK)
        self.gauss_filter_width = 1  # filter width for gaussian filter (GFSK)

        self.parameters = array.array(
            "f", [0, 100]
        )  # Freq, Amplitude (0..100%) or Phase (0..360)

    def __eq__(self, other):
        return (
            self.carrier_freq_hz == other.carrier_freq_hz
            and self.carrier_amplitude == other.carrier_amplitude
            and self.carrier_phase_deg == other.carrier_phase_deg
            and self.name == other.name
            and self.modulation_type == other.modulation_type
            and self.samples_per_symbol == other.samples_per_symbol
            and self.bits_per_symbol == other.bits_per_symbol
            and self.sample_rate == other.sample_rate
            and self.parameters == other.parameters
        )

    @staticmethod
    def get_dtype():
        if Modulator.FORCE_DTYPE is not None:
            return Modulator.FORCE_DTYPE

        dtype_str = settings.read("modulation_dtype", "float32", str)
        if dtype_str == "int8":
            return np.int8
        elif dtype_str == "int16":
            return np.int16
        else:
            return np.float32

    @property
    def modulation_type(self) -> str:
        return self.__modulation_type

    @modulation_type.setter
    def modulation_type(self, value):
        try:
            # legacy support when modulation type was saved as int index
            self.__modulation_type = self.MODULATION_TYPES[int(value)]
        except (ValueError, IndexError):
            self.__modulation_type = value

    @property
    def is_binary_modulation(self):
        return self.bits_per_symbol == 1

    @property
    def is_amplitude_based(self):
        return "ASK" in self.modulation_type

    @property
    def is_frequency_based(self):
        return "FSK" in self.modulation_type

    @property
    def is_phase_based(self):
        return "PSK" in self.modulation_type

    @property
    def bits_per_symbol(self):
        return self.__bits_per_symbol

    @bits_per_symbol.setter
    def bits_per_symbol(self, value):
        value = int(value)
        if value != self.bits_per_symbol:
            self.__bits_per_symbol = value
            self.parameters = array.array("f", [0] * self.modulation_order)

    @property
    def modulation_order(self):
        return 2**self.bits_per_symbol

    @property
    def parameter_type_str(self) -> str:
        if self.is_amplitude_based:
            return "Amplitudes in %:"
        if self.is_frequency_based:
            return "Frequencies in Hz:"
        if self.is_phase_based:
            return "Phases in degree:"

        return "Unknown Modulation Type (This should not happen...)"

    @property
    def parameters_string(self) -> str:
        if self.is_amplitude_based:
            return "/".join(map(str, map(int, self.parameters)))
        elif self.is_frequency_based:
            return "/".join(map(Formatter.big_value_with_suffix, self.parameters))
        elif self.is_phase_based:
            return "/".join(map(str, map(int, self.parameters)))
        else:
            raise ValueError("")

    @property
    def sample_rate(self):
        if self.__sample_rate is not None:
            return self.__sample_rate
        else:
            return self.default_sample_rate

    @sample_rate.setter
    def sample_rate(self, value):
        self.__sample_rate = value

    @property
    def display_bits(self) -> str:
        return "".join(["1" if bit else "0" for bit in self.data])

    @display_bits.setter
    def display_bits(self, value: str):
        self.data = [True if bit == "1" else False for bit in value]

    @property
    def carrier_frequency_str(self):
        return self.get_value_with_suffix(self.carrier_freq_hz, unit="Hz")

    @property
    def carrier_phase_str(self):
        return self.get_value_with_suffix(self.carrier_phase_deg, unit="°")

    @property
    def samples_per_symbol_str(self):
        return self.get_value_with_suffix(self.samples_per_symbol, unit="")

    @property
    def sample_rate_str(self):
        return self.get_value_with_suffix(self.sample_rate)

    @property
    def modulation_type_verbose(self):
        return self.MODULATION_TYPES_VERBOSE[self.modulation_type]

    @property
    def carrier_data(self):
        num_samples = len(self.display_bits) * self.samples_per_symbol
        carrier_phase_rad = self.carrier_phase_deg * (np.pi / 180)
        t = (np.arange(0, num_samples) / self.sample_rate).astype(np.float32)
        arg = (2 * np.pi * self.carrier_freq_hz * t + carrier_phase_rad).astype(
            np.float32
        )
        y = self.carrier_amplitude * np.sin(arg)  # type: np.ndarray

        return y.astype(np.float32)

    @property
    def data_scene(self) -> QGraphicsScene:
        n = self.samples_per_symbol * len(self.display_bits)
        y = np.ones(n, dtype=np.float32)

        for i, bit in enumerate(self.display_bits):
            if bit == "0":
                y[
                    i * self.samples_per_symbol : (i + 1) * self.samples_per_symbol
                ] = -1.0

        x = np.arange(0, n).astype(np.int64)

        scene = ZoomableScene()
        scene.setSceneRect(0, -1.25, n, 2.5)
        scene.setBackgroundBrush(settings.BGCOLOR)
        scene.addLine(0, 0, n, 0, QPen(settings.AXISCOLOR, 0))

        path = path_creator.array_to_QPath(x, y)
        scene.addPath(path, QPen(settings.LINECOLOR, 0))

        return scene

    def modulate(self, data=None, pause=0, start=0, dtype=None) -> IQArray:
        assert pause >= 0
        if data is None:
            data = self.data
        else:
            self.data = data

        if isinstance(data, str):
            data = array.array("B", map(int, data))
        elif isinstance(data, list):
            data = array.array("B", data)

        if len(data) == 0:
            return IQArray(None, np.float32, 0)

        dtype = dtype or self.get_dtype()
        a = self.carrier_amplitude * IQArray.min_max_for_dtype(dtype)[1]

        parameters = self.parameters
        if self.modulation_type == "ASK":
            parameters = array.array("f", [a * p / 100 for p in parameters])
        elif self.modulation_type == "PSK":
            parameters = array.array("f", [p * (math.pi / 180) for p in parameters])

        result = signal_functions.modulate_c(
            data,
            self.samples_per_symbol,
            self.modulation_type,
            parameters,
            self.bits_per_symbol,
            a,
            self.carrier_freq_hz,
            self.carrier_phase_deg * (np.pi / 180),
            self.sample_rate,
            pause,
            start,
            dtype,
            self.gauss_bt,
            self.gauss_filter_width,
        )
        return IQArray(result)

    def get_default_parameters(self) -> array.array:
        if self.is_amplitude_based:
            parameters = np.linspace(0, 100, self.modulation_order, dtype=np.float32)
        elif self.is_frequency_based:
            parameters = []
            for i in range(self.modulation_order):
                parameters.append(
                    (i + 1) * self.carrier_freq_hz / self.modulation_order
                )
        elif self.is_phase_based:
            step = 360 / self.modulation_order
            parameters = np.arange(step / 2, 360, step) - 180
            if self.modulation_type == "OQPSK":
                parameters = parameters[
                    (self.__get_gray_code_indices(self.modulation_order))
                ]
        else:
            return None

        return array.array("f", parameters)

    @staticmethod
    def __get_gray_code_indices(n: int):
        result = []
        for i in range(0, n):
            result.append(i ^ (i >> 1))
        return result

    def to_xml(self, index: int) -> ET.Element:
        root = ET.Element("modulator")

        for attr, val in vars(self).items():
            if attr not in (
                "data",
                "_Modulator__sample_rate",
                "_Modulator__modulation_type",
                "_Modulator__bits_per_symbol",
                "default_sample_rate",
                "parameters",
            ):
                root.set(attr, str(val))

        root.set("sample_rate", str(self.__sample_rate))
        root.set("modulation_type", self.__modulation_type)
        root.set("index", str(index))
        root.set("parameters", ",".join(map(str, self.parameters)))
        root.set("bits_per_symbol", str(self.bits_per_symbol))

        return root

    def estimate_carrier_frequency(self, signal, protocol) -> int or None:
        if len(protocol.messages) == 0:
            return None

        # Take the first message for detection
        start, num_samples = protocol.get_samplepos_of_bitseq(0, 0, 0, 999999, False)
        # Avoid too large arrays
        if num_samples > 1e6:
            num_samples = int(1e6)

        return signal.estimate_frequency(start, start + num_samples, self.sample_rate)

    @staticmethod
    def from_xml(tag: ET.Element):
        result = Modulator("")
        for attrib, value in sorted(tag.attrib.items()):
            if attrib == "index":
                continue
            elif attrib == "name" or attrib == "modulation_type":
                setattr(result, attrib, str(value))
            elif attrib == "samples_per_bit" or attrib == "samples_per_symbol":
                # samples_per_bit as legacy support for older project files
                result.samples_per_symbol = Formatter.str2val(value, int, 100)
            elif attrib == "sample_rate":
                result.sample_rate = (
                    Formatter.str2val(value, float, 1e6) if value != "None" else None
                )
            elif attrib == "param_for_zero":
                result.parameters[0] = Formatter.str2val(value, float, 0)  # legacy
            elif attrib == "param_for_one":
                result.parameters[1] = Formatter.str2val(value, float, 100)  # legacy
            elif attrib == "bits_per_symbol":
                result.bits_per_symbol = Formatter.str2val(value, int, 1)
            elif attrib == "parameters":
                try:
                    result.parameters = array.array("f", map(float, value.split(",")))
                except ValueError:
                    continue
            elif not attrib.startswith("_Modulator__"):
                setattr(result, attrib, Formatter.str2val(value, float, 1))
        return result

    @staticmethod
    def modulators_to_xml_tag(modulators: list) -> ET.Element:
        modulators_tag = ET.Element("modulators")
        for i, modulator in enumerate(modulators):
            modulators_tag.append(modulator.to_xml(i))
        return modulators_tag

    @staticmethod
    def modulators_from_xml_tag(xml_tag: ET.Element) -> list:
        if xml_tag is None:
            return []

        if xml_tag.tag != "modulators":
            xml_tag = xml_tag.find("modulators")

        if xml_tag is None:
            return []

        result = []
        for mod_tag in xml_tag.iter("modulator"):
            result.append(Modulator.from_xml(mod_tag))
        return result

    @staticmethod
    def get_value_with_suffix(value, unit=""):
        decimal_point = locale.localeconv()["decimal_point"]

        if abs(value) >= 10**9:
            target_val, suffix = value / 10**9, "G"
        elif abs(value) >= 10**6:
            target_val, suffix = value / 10**6, "M"
        elif abs(value) >= 10**3:
            target_val, suffix = value / 10**3, "k"
        else:
            target_val, suffix = value, ""

        return (
            locale.format_string("%.3f", target_val).rstrip("0").rstrip(decimal_point)
            + suffix
            + unit
        )



# -*- coding: utf-8 -*-

#
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_DialogLabels(object):
    def setupUi(self, DialogLabels):
        DialogLabels.setObjectName("DialogLabels")
        DialogLabels.resize(714, 463)
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(DialogLabels)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.splitter = QtWidgets.QSplitter(DialogLabels)
        self.splitter.setStyleSheet(
            "QSplitter::handle:vertical {\n"
            "margin: 4px 0px;\n"
            "    background-color: qlineargradient(x1:0, y1:0, x2:1, y2:0, \n"
            "stop:0 rgba(255, 255, 255, 0), \n"
            "stop:0.5 rgba(100, 100, 100, 100), \n"
            "stop:1 rgba(255, 255, 255, 0));\n"
            "    image: url(:/icons/icons/splitter_handle_horizontal.svg);\n"
            "}"
        )
        self.splitter.setOrientation(QtCore.Qt.Vertical)
        self.splitter.setHandleWidth(6)
        self.splitter.setChildrenCollapsible(False)
        self.splitter.setObjectName("splitter")
        self.groupBoxSettings = QtWidgets.QGroupBox(self.splitter)
        self.groupBoxSettings.setObjectName("groupBoxSettings")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.groupBoxSettings)
        self.verticalLayout.setObjectName("verticalLayout")
        self.tblViewProtoLabels = ProtocolLabelTableView(self.groupBoxSettings)
        self.tblViewProtoLabels.setSelectionMode(
            QtWidgets.QAbstractItemView.ExtendedSelection
        )
        self.tblViewProtoLabels.setSelectionBehavior(
            QtWidgets.QAbstractItemView.SelectItems
        )
        self.tblViewProtoLabels.setVerticalScrollMode(
            QtWidgets.QAbstractItemView.ScrollPerPixel
        )
        self.tblViewProtoLabels.setHorizontalScrollMode(
            QtWidgets.QAbstractItemView.ScrollPerPixel
        )
        self.tblViewProtoLabels.setShowGrid(False)
        self.tblViewProtoLabels.setObjectName("tblViewProtoLabels")
        self.verticalLayout.addWidget(self.tblViewProtoLabels)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.label = QtWidgets.QLabel(self.groupBoxSettings)
        font = QtGui.QFont()
        font.setUnderline(False)
        self.label.setFont(font)
        self.label.setObjectName("label")
        self.horizontalLayout_2.addWidget(self.label)
        self.cbProtoView = QtWidgets.QComboBox(self.groupBoxSettings)
        self.cbProtoView.setObjectName("cbProtoView")
        self.cbProtoView.addItem("")
        self.cbProtoView.addItem("")
        self.cbProtoView.addItem("")
        self.horizontalLayout_2.addWidget(self.cbProtoView)
        spacerItem = QtWidgets.QSpacerItem(
            40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum
        )
        self.horizontalLayout_2.addItem(spacerItem)
        self.verticalLayout.addLayout(self.horizontalLayout_2)
        self.groupBoxAdvancedSettings = QtWidgets.QGroupBox(self.splitter)
        self.groupBoxAdvancedSettings.setObjectName("groupBoxAdvancedSettings")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.groupBoxAdvancedSettings)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.tabWidgetAdvancedSettings = QtWidgets.QTabWidget(
            self.groupBoxAdvancedSettings
        )
        self.tabWidgetAdvancedSettings.setObjectName("tabWidgetAdvancedSettings")
        self.verticalLayout_2.addWidget(self.tabWidgetAdvancedSettings)
        self.verticalLayout_3.addWidget(self.splitter)
        self.btnConfirm = QtWidgets.QPushButton(DialogLabels)
        self.btnConfirm.setObjectName("btnConfirm")
        self.verticalLayout_3.addWidget(self.btnConfirm)

        self.retranslateUi(DialogLabels)

    def retranslateUi(self, DialogLabels):
        _translate = QtCore.QCoreApplication.translate
        DialogLabels.setWindowTitle(
            _translate("DialogLabels", "Manage Protocol Labels")
        )
        self.groupBoxSettings.setTitle(
            _translate("DialogLabels", "Protocol Label Settings")
        )
        self.label.setText(
            _translate("DialogLabels", "Start/End values refer to view type:")
        )
        self.cbProtoView.setItemText(0, _translate("DialogLabels", "Bits"))
        self.cbProtoView.setItemText(1, _translate("DialogLabels", "Hex"))
        self.cbProtoView.setItemText(2, _translate("DialogLabels", "ASCII"))
        self.groupBoxAdvancedSettings.setTitle(
            _translate("DialogLabels", "Advanced Settings")
        )
        self.btnConfirm.setText(_translate("DialogLabels", "Confirm"))


from urh.ui.views.ProtocolLabelTableView import ProtocolLabelTableView
from . import urh_rc



from PyQt5.QtCore import Qt, QModelIndex, QAbstractTableModel, pyqtSignal
from PyQt5.QtGui import QFont

from urh.signalprocessing.MessageType import MessageType


class MessageTypeTableModel(QAbstractTableModel):
    message_type_visibility_changed = pyqtSignal(MessageType)
    message_type_removed = pyqtSignal(MessageType)
    message_type_name_edited = pyqtSignal(str)
    header_labels = ["Name", "Edit"]

    def __init__(self, message_types: list, parent=None):
        super().__init__(parent)
        self.message_types = message_types  # type: list[MessageType]

        self.selected_message_type_indices = set()

    def get_num_active_rules_of_message_type_at(self, index: int) -> int:
        try:
            if self.message_types[index].assigned_by_ruleset:
                return len(self.message_types[index].ruleset)
            else:
                return 0
        except IndexError:
            return 0

    def rowCount(self, QModelIndex_parent=None, *args, **kwargs):
        return len(self.message_types)

    def columnCount(self, parent: QModelIndex = None, *args, **kwargs):
        return len(self.header_labels)

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        if role == Qt.DisplayRole and orientation == Qt.Horizontal:
            return self.header_labels[section]
        return super().headerData(section, orientation, role)

    def update(self):
        self.beginResetModel()
        self.endResetModel()

    def data(self, index, role=Qt.DisplayRole):
        row = index.row()
        if not index.isValid() or row >= len(self.message_types):
            return

        message_type = self.message_types[row]

        if role == Qt.DisplayRole:
            if index.column() == 0:
                return message_type.name
            elif index.column() == 1:
                return ""
        elif role == Qt.CheckStateRole:
            if index.column() == 0:
                return message_type.show
            elif index.column() == 1:
                return None
        elif role == Qt.EditRole:
            if index.column() == 0:
                return message_type.name
        elif role == Qt.FontRole and index.column() == 0:
            font = QFont()
            font.setBold(index.row() in self.selected_message_type_indices)
            return font

    def setData(self, index: QModelIndex, value, role=Qt.DisplayRole):
        if role == Qt.CheckStateRole:
            if index.column() == 0:
                message_type = self.message_types[index.row()]
                message_type.show = value
                self.message_type_visibility_changed.emit(message_type)
        elif role == Qt.EditRole:
            if index.column() == 0 and value:
                message_type = self.message_types[index.row()]
                message_type.name = value
                self.message_type_name_edited.emit(value)

        return True

    def delete_message_type_at(self, index: int):
        try:
            message_type = self.message_types[index]
            self.message_types.remove(message_type)
            self.message_type_removed.emit(message_type)
        except IndexError:
            pass

    def delete_message_types_at(self, start: int, end: int):
        for row in range(end, start - 1, -1):
            self.delete_message_type_at(row)

    def flags(self, index):
        return (
            Qt.ItemIsEnabled
            | Qt.ItemIsSelectable
            | Qt.ItemIsUserCheckable
            | Qt.ItemIsEditable
        )



# -*- coding: utf-8 -*-

#
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_FramePlugins(object):
    def setupUi(self, FramePlugins):
        FramePlugins.setObjectName("FramePlugins")
        FramePlugins.resize(806, 683)
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(FramePlugins)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout()
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.label = QtWidgets.QLabel(FramePlugins)
        self.label.setObjectName("label")
        self.verticalLayout_2.addWidget(self.label)
        self.listViewPlugins = QtWidgets.QListView(FramePlugins)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.MinimumExpanding
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.listViewPlugins.sizePolicy().hasHeightForWidth()
        )
        self.listViewPlugins.setSizePolicy(sizePolicy)
        self.listViewPlugins.setObjectName("listViewPlugins")
        self.verticalLayout_2.addWidget(self.listViewPlugins)
        self.horizontalLayout.addLayout(self.verticalLayout_2)
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.label_2 = QtWidgets.QLabel(FramePlugins)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_2.sizePolicy().hasHeightForWidth())
        self.label_2.setSizePolicy(sizePolicy)
        self.label_2.setObjectName("label_2")
        self.verticalLayout.addWidget(self.label_2)
        self.txtEditPluginDescription = QtWidgets.QTextEdit(FramePlugins)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.txtEditPluginDescription.sizePolicy().hasHeightForWidth()
        )
        self.txtEditPluginDescription.setSizePolicy(sizePolicy)
        self.txtEditPluginDescription.setReadOnly(True)
        self.txtEditPluginDescription.setPlaceholderText("")
        self.txtEditPluginDescription.setObjectName("txtEditPluginDescription")
        self.verticalLayout.addWidget(self.txtEditPluginDescription)
        self.horizontalLayout.addLayout(self.verticalLayout)
        self.verticalLayout_3.addLayout(self.horizontalLayout)
        self.groupBoxSettings = QtWidgets.QGroupBox(FramePlugins)
        sizePolicy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Expanding
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.groupBoxSettings.sizePolicy().hasHeightForWidth()
        )
        self.groupBoxSettings.setSizePolicy(sizePolicy)
        self.groupBoxSettings.setMaximumSize(QtCore.QSize(16777215, 16777215))
        self.groupBoxSettings.setObjectName("groupBoxSettings")
        self.verticalLayout_3.addWidget(self.groupBoxSettings)

        self.retranslateUi(FramePlugins)

    def retranslateUi(self, FramePlugins):
        _translate = QtCore.QCoreApplication.translate
        FramePlugins.setWindowTitle(_translate("FramePlugins", "Plugins"))
        self.label.setText(_translate("FramePlugins", "Available Plugins"))
        self.label_2.setText(_translate("FramePlugins", "Description"))
        self.groupBoxSettings.setTitle(_translate("FramePlugins", "Settings"))



cdef extern from "mirsdrapi-rsp.h":
    ctypedef struct mir_sdr_DeviceT:
       char *SerNo
       char *DevNm
       unsigned char hwVer
       unsigned char devAvail

    ctypedef enum mir_sdr_Bw_MHzT:
      mir_sdr_BW_Undefined = 0
      mir_sdr_BW_0_200     = 200
      mir_sdr_BW_0_300     = 300
      mir_sdr_BW_0_600     = 600
      mir_sdr_BW_1_536     = 1536
      mir_sdr_BW_5_000     = 5000
      mir_sdr_BW_6_000     = 6000
      mir_sdr_BW_7_000     = 7000
      mir_sdr_BW_8_000     = 8000

    ctypedef enum mir_sdr_If_kHzT:
      mir_sdr_IF_Undefined = -1
      mir_sdr_IF_Zero      = 0
      mir_sdr_IF_0_450     = 450
      mir_sdr_IF_1_620     = 1620
      mir_sdr_IF_2_048     = 2048

    ctypedef enum mir_sdr_ErrT:
      mir_sdr_Success            = 0
      mir_sdr_Fail               = 1
      mir_sdr_InvalidParam       = 2
      mir_sdr_OutOfRange         = 3
      mir_sdr_GainUpdateError    = 4
      mir_sdr_RfUpdateError      = 5
      mir_sdr_FsUpdateError      = 6
      mir_sdr_HwError            = 7
      mir_sdr_AliasingError      = 8
      mir_sdr_AlreadyInitialised = 9
      mir_sdr_NotInitialised     = 10
      mir_sdr_NotEnabled         = 11
      mir_sdr_HwVerError         = 12
      mir_sdr_OutOfMemError      = 13

    ctypedef enum mir_sdr_TransferModeT:
      mir_sdr_ISOCH = 0
      mir_sdr_BULK  = 1

    ctypedef enum mir_sdr_ReasonForReinitT:
      mir_sdr_CHANGE_NONE    = 0x00
      mir_sdr_CHANGE_GR      = 0x01
      mir_sdr_CHANGE_FS_FREQ = 0x02
      mir_sdr_CHANGE_RF_FREQ = 0x04
      mir_sdr_CHANGE_BW_TYPE = 0x08
      mir_sdr_CHANGE_IF_TYPE = 0x10
      mir_sdr_CHANGE_LO_MODE = 0x20
      mir_sdr_CHANGE_AM_PORT = 0x40

    ctypedef enum mir_sdr_LoModeT:
      mir_sdr_LO_Undefined = 0
      mir_sdr_LO_Auto      = 1
      mir_sdr_LO_120MHz    = 2
      mir_sdr_LO_144MHz    = 3
      mir_sdr_LO_168MHz    = 4

    ctypedef enum mir_sdr_BandT:
      mir_sdr_BAND_AM_LO   = 0
      mir_sdr_BAND_AM_MID  = 1
      mir_sdr_BAND_AM_HI   = 2
      mir_sdr_BAND_VHF     = 3
      mir_sdr_BAND_3       = 4
      mir_sdr_BAND_X       = 5
      mir_sdr_BAND_4_5     = 6
      mir_sdr_BAND_L       = 7

    ctypedef enum mir_sdr_SetGrModeT:
      mir_sdr_USE_SET_GR                = 0
      mir_sdr_USE_SET_GR_ALT_MODE       = 1
      mir_sdr_USE_RSP_SET_GR            = 2

    ctypedef enum mir_sdr_RSPII_BandT:
      mir_sdr_RSPII_BAND_UNKNOWN = 0
      mir_sdr_RSPII_BAND_AM_LO   = 1
      mir_sdr_RSPII_BAND_AM_MID  = 2
      mir_sdr_RSPII_BAND_AM_HI   = 3
      mir_sdr_RSPII_BAND_VHF     = 4
      mir_sdr_RSPII_BAND_3       = 5
      mir_sdr_RSPII_BAND_X_LO    = 6
      mir_sdr_RSPII_BAND_X_MID   = 7
      mir_sdr_RSPII_BAND_X_HI    = 8
      mir_sdr_RSPII_BAND_4_5     = 9
      mir_sdr_RSPII_BAND_L       = 10

    ctypedef enum mir_sdr_RSPII_AntennaSelectT:
      mir_sdr_RSPII_ANTENNA_A = 5
      mir_sdr_RSPII_ANTENNA_B = 6

    ctypedef enum mir_sdr_AgcControlT:
      mir_sdr_AGC_DISABLE  = 0
      mir_sdr_AGC_100HZ    = 1
      mir_sdr_AGC_50HZ     = 2
      mir_sdr_AGC_5HZ      = 3

    ctypedef enum mir_sdr_GainMessageIdT:
      mir_sdr_GAIN_MESSAGE_START_ID  = 0x80000000
      mir_sdr_ADC_OVERLOAD_DETECTED  = mir_sdr_GAIN_MESSAGE_START_ID + 1
      mir_sdr_ADC_OVERLOAD_CORRECTED = mir_sdr_GAIN_MESSAGE_START_ID + 2

    ctypedef enum mir_sdr_MinGainReductionT:
      mir_sdr_EXTENDED_MIN_GR = 0
      mir_sdr_NORMAL_MIN_GR   = 20

    ctypedef struct mir_sdr_GainValuesT:
       float curr;
       float max;
       float min;

    ctypedef void mir_sdr_StreamCallback_t(short *xi, short *xq, unsigned int firstSampleNum, int grChanged, int rfChanged, int fsChanged, unsigned int numSamples, unsigned int reset, unsigned int hwRemoved, void *cbContext)
    ctypedef void mir_sdr_GainChangeCallback_t(unsigned int gRdB, unsigned int lnaGRdB, void *cbContext)

    mir_sdr_ErrT mir_sdr_ReadPacket(short *xi, short *xq, unsigned int *firstSampleNum, int *grChanged, int *rfChanged, int *fsChanged)
    mir_sdr_ErrT mir_sdr_SetRf(double drfHz, int abs, int syncUpdate)
    mir_sdr_ErrT mir_sdr_SetFs(double dfsHz, int abs, int syncUpdate, int reCal)
    mir_sdr_ErrT mir_sdr_SetGr(int gRdB, int abs, int syncUpdate)
    mir_sdr_ErrT mir_sdr_SetGrParams(int minimumGr, int lnaGrThreshold)
    mir_sdr_ErrT mir_sdr_SetDcMode(int dcCal, int speedUp)
    mir_sdr_ErrT mir_sdr_SetDcTrackTime(int trackTime)
    mir_sdr_ErrT mir_sdr_SetSyncUpdateSampleNum(unsigned int sampleNum)
    mir_sdr_ErrT mir_sdr_SetSyncUpdatePeriod(unsigned int period)
    mir_sdr_ErrT mir_sdr_ApiVersion(float *version)
    mir_sdr_ErrT mir_sdr_ResetUpdateFlags(int resetGainUpdate, int resetRfUpdate, int resetFsUpdate)
    mir_sdr_ErrT mir_sdr_SetParam(unsigned int id, unsigned int value)
    mir_sdr_ErrT mir_sdr_SetPpm(double ppm)
    mir_sdr_ErrT mir_sdr_SetGrAltMode(int *gRidx, int LNAstate, int *gRdBsystem, int abs, int syncUpdate)
    mir_sdr_ErrT mir_sdr_DCoffsetIQimbalanceControl(unsigned int DCenable, unsigned int IQenable)
    mir_sdr_ErrT mir_sdr_DecimateControl(unsigned int enable, unsigned int decimationFactor, unsigned int wideBandSignal)

    mir_sdr_ErrT mir_sdr_StreamInit(int *gRdB, double fsMHz, double rfMHz, mir_sdr_Bw_MHzT bwType, mir_sdr_If_kHzT ifType, int LNAstate, int *gRdBsystem, mir_sdr_SetGrModeT setGrMode, int *samplesPerPacket, mir_sdr_StreamCallback_t StreamCbFn, mir_sdr_GainChangeCallback_t GainChangeCbFn, void *cbContext)
    mir_sdr_ErrT mir_sdr_StreamUninit()
    mir_sdr_ErrT mir_sdr_Reinit(int *gRdB, double fsMHz, double rfMHz, mir_sdr_Bw_MHzT bwType, mir_sdr_If_kHzT ifType, mir_sdr_LoModeT loMode, int LNAstate, int *gRdBsystem, mir_sdr_SetGrModeT setGrMode, int *samplesPerPacket, mir_sdr_ReasonForReinitT reasonForReinit)
    mir_sdr_ErrT mir_sdr_DebugEnable(unsigned int enable)

    mir_sdr_ErrT mir_sdr_GetDevices(mir_sdr_DeviceT *devices, unsigned int *numDevs, unsigned int maxDevs)
    mir_sdr_ErrT mir_sdr_SetDeviceIdx(unsigned int idx)
    mir_sdr_ErrT mir_sdr_ReleaseDeviceIdx()
    mir_sdr_ErrT mir_sdr_GetHwVersion(unsigned char *ver)
    mir_sdr_ErrT mir_sdr_RSPII_AntennaControl(mir_sdr_RSPII_AntennaSelectT select)
    mir_sdr_ErrT mir_sdr_RSPII_ExternalReferenceControl(unsigned int output_enable)
    mir_sdr_ErrT mir_sdr_RSPII_BiasTControl(unsigned int enable)
    mir_sdr_ErrT mir_sdr_RSPII_RfNotchEnable(unsigned int enable)

    mir_sdr_ErrT mir_sdr_RSP_SetGr(int gRdB, int LNAstate, int abs, int syncUpdate)
    mir_sdr_ErrT mir_sdr_RSP_SetGrLimits(mir_sdr_MinGainReductionT minGr)

    mir_sdr_ErrT mir_sdr_AmPortSelect(int port)



from libcpp cimport bool

ctypedef unsigned int  uint32_t
ctypedef unsigned long long uint64_t

cdef extern from "lime/LimeSuite.h":
    ctypedef double float_type
    int LMS_SUCCESS = 0

    ctypedef void lms_device_t
    ctypedef char lms_info_str_t[256]

    int LMS_GetDeviceList(lms_info_str_t *dev_list)
    int LMS_Open(lms_device_t ** device, lms_info_str_t info, void*args)
    int LMS_Close(lms_device_t *device)

    bool LMS_CH_TX = True
    bool LMS_CH_RX = False

    ctypedef struct lms_range_t:
        float_type min  # Minimum allowed value
        float_type max  # Minimum allowed value
        float_type step  # Minimum value step

    ctypedef enum lms_testsig_t:
        LMS_TESTSIG_NONE = 0  # Disable test signals. Return to normal operation
        LMS_TESTSIG_NCODIV8  # Test signal from NCO half scale
        LMS_TESTSIG_NCODIV4  # Test signal from NCO half scale
        LMS_TESTSIG_NCODIV8F  # Test signal from NCO full scale
        LMS_TESTSIG_NCODIV4F  # Test signal from NCO full scale
        LMS_TESTSIG_DC  # DC test signal

    int LMS_Init(lms_device_t *device)
    int LMS_Reset(lms_device_t *device)
    int LMS_Synchronize(lms_device_t *dev, bool to_chip)
    int LMS_GetNumChannels(lms_device_t *device, bool dir_tx)
    int LMS_EnableChannel(lms_device_t *device, bool dir_tx, size_t chan, bool enabled)

    int LMS_SaveConfig(lms_device_t *device, const char *filename)
    int LMS_LoadConfig(lms_device_t *device, const char *filename)

    int LMS_SetSampleRate(lms_device_t *device, float_type rate, size_t oversample)
    int LMS_GetSampleRate(lms_device_t *device, bool dir_tx, size_t chan, float_type *host_Hz, float_type *rf_Hz)
    int LMS_GetSampleRateRange(lms_device_t *device, bool dir_tx, lms_range_t *range)

    int LMS_SetLOFrequency(lms_device_t *device, bool dir_tx, size_t chan, float_type frequency)
    int LMS_GetLOFrequency(lms_device_t *device, bool dir_tx, size_t chan, float_type *frequency)
    int LMS_GetLOFrequencyRange(lms_device_t *device, bool dir_tx, lms_range_t *range)

    int LMS_SetNormalizedGain(lms_device_t *device, bool dir_tx, size_t chan, float_type gain)
    int LMS_GetNormalizedGain(lms_device_t *device, bool dir_tx, size_t chan, float_type *gain)

    int LMS_SetLPFBW(lms_device_t *device, bool dir_tx, size_t chan, float_type bandwidth)
    int LMS_GetLPFBW(lms_device_t *device, bool dir_tx, size_t chan, float_type *bandwidth)
    int LMS_GetLPFBWRange(lms_device_t *device, bool dir_tx, lms_range_t *range)
    int LMS_SetLPF(lms_device_t *device, bool dir_tx, size_t chan, bool enabled)
    int LMS_SetGFIRLPF(lms_device_t *device, bool dir_tx, size_t chan, bool enabled, float_type bandwidth)

    int LMS_Calibrate(lms_device_t *device, bool dir_tx, size_t chan, double bw, unsigned flags)

    int LMS_SetNCOFrequency(lms_device_t *device, bool dir_tx, size_t chan, const float_type *freq, float_type pho)
    int LMS_GetNCOFrequency(lms_device_t *device, bool dir_tx, size_t chan, float_type *freq, float_type *pho)

    int LMS_GetClockFreq(lms_device_t *dev, size_t clk_id, float_type *freq)
    int LMS_SetClockFreq(lms_device_t *dev, size_t clk_id, float_type freq)

    ctypedef char lms_name_t[16]
    int LMS_GetAntennaList(lms_device_t *device, bool dir_tx, size_t chan, lms_name_t *list)
    int LMS_SetAntenna(lms_device_t *device, bool dir_tx, size_t chan, size_t index)
    int LMS_GetAntenna(lms_device_t *device, bool dir_tx, size_t chan)
    int LMS_GetAntennaBW(lms_device_t *device, bool dir_tx, size_t chan, size_t index, lms_range_t *range)

    int LMS_GetChipTemperature(lms_device_t *dev, size_t ind, float_type *temp)

    ctypedef struct lms_stream_meta_t:
        # Timestamp is a value of HW counter with a tick based on sample rate.
        # In RX: time when the first sample in the returned buffer was received
        # In TX: time when the first sample in the submitted buffer should be send
        uint64_t timestamp

        # In TX: wait for the specified HW timestamp before broadcasting data over the air
        # In RX: wait for the specified HW timestamp before starting to receive samples
        bool waitForTimestamp

        # Indicates the end of send/receive transaction. Discards data remainder
        # in buffer (if there is any) in RX or flushes transfer buffer in TX (even
        # if the buffer is not full yet)
        bool flushPartialPacket

    ctypedef enum dataFmt_t:
        LMS_FMT_F32 "lms_stream_t::LMS_FMT_F32" = 0
        LMS_FMT_I16 "lms_stream_t::LMS_FMT_I16"
        LMS_FMT_I12 "lms_stream_t::LMS_FMT_I12"

    ctypedef struct lms_stream_t:
        # Stream handle. Should not be modified manually. Assigned by LMS_SetupStream()
        size_t handle

        # Indicates whether stream is TX (true) or RX (false)
        bool isTx

        # Channel number. Starts at 0.
        uint32_t channel

        # FIFO size (in samples) used by stream.
        uint32_t fifoSize

        # Parameter for controlling configuration bias toward low latency or high data throughput range [0,1.0].
        # 0 - lowest latency, usually results in lower throughput
        # 1 - higher throughput, usually results in higher latency
        float throughputVsLatency

        dataFmt_t dataFmt

    ctypedef struct lms_stream_status_t:
        bool active  # Indicates whether the stream is currently active
        uint32_t fifoFilledCount  # Number of samples in FIFO buffer
        uint32_t fifoSize  # Size of FIFO buffer
        uint32_t underrun  # FIFO underrun count
        uint32_t overrun  # FIFO overrun count
        uint32_t droppedPackets  # Number of dropped packets by HW
        float_type sampleRate  # Sampling rate of the stream
        float_type linkRate  # Combined data rate of all stream of the same direction (TX or RX)
        uint64_t timestamp  # Current HW timestamp

    int LMS_SetupStream(lms_device_t *device, lms_stream_t *stream)
    int LMS_DestroyStream(lms_device_t *device, lms_stream_t *stream)
    int LMS_StartStream(lms_stream_t *stream)
    int LMS_StopStream(lms_stream_t *conf)
    int LMS_GetStreamStatus(lms_stream_t *stream, lms_stream_status_t*status)
    int LMS_RecvStream(lms_stream_t *stream, void *samples, size_t sample_count, lms_stream_meta_t *meta,
                       unsigned timeout_ms)
    int LMS_SendStream(lms_stream_t *stream, const void *samples, size_t sample_count, const lms_stream_meta_t *meta,
                       unsigned timeout_ms)

    const char* LMS_GetLastErrorMessage()



import numpy as np

from urh import settings
from urh.dev.VirtualDevice import Mode, VirtualDevice
from urh.util.RingBuffer import RingBuffer


class EndlessSender(object):
    """
    Enter endless send mode for a device and send data if data gets pushed to ringbuffer.
    """

    def __init__(self, backend_handler, name: str):
        self.__device = VirtualDevice(
            backend_handler=backend_handler, name=name, mode=Mode.send
        )
        self.ringbuffer = RingBuffer(
            int(settings.CONTINUOUS_BUFFER_SIZE_MB * 10**6) // 8,
            self.__device.data_type,
        )
        self.__device.continuous_send_ring_buffer = self.ringbuffer
        self.__device.is_send_continuous = True

    @property
    def device(self) -> VirtualDevice:
        return self.__device

    @device.setter
    def device(self, value: VirtualDevice):
        self.__device = value
        self.__device.is_send_continuous = True
        self.ringbuffer = RingBuffer(
            int(settings.CONTINUOUS_BUFFER_SIZE_MB * 10**6) // 8,
            self.__device.data_type,
        )
        self.__device.continuous_send_ring_buffer = self.ringbuffer

    @property
    def device_name(self) -> str:
        return self.device.name

    @device_name.setter
    def device_name(self, value: str):
        if value != self.device_name:
            self.device = VirtualDevice(
                backend_handler=self.device.backend_handler, name=value, mode=Mode.send
            )

    def start(self):
        self.device.num_sending_repeats = 0
        self.device.start()

    def stop(self):
        self.device.stop("EndlessSender stopped.")

    def push_data(self, data: np.ndarray):
        self.ringbuffer.push(data)


if __name__ == "__main__":
    from urh.dev.BackendHandler import BackendHandler
    from urh.signalprocessing.Message import Message
    from urh.signalprocessing.MessageType import MessageType
    from urh.signalprocessing.Modulator import Modulator
    from urh.util.Logger import logger
    import time

    endless_sender = EndlessSender(BackendHandler(), "HackRF")
    msg = Message(
        [1, 0] * 16
        + [1, 1, 0, 0] * 8
        + [0, 0, 1, 1] * 8
        + [1, 0, 1, 1, 1, 0, 0, 1, 1, 1] * 4,
        0,
        MessageType("empty_message_type"),
    )
    modulator = Modulator("test_modulator")
    modulator.samples_per_symbol = 1000
    modulator.carrier_freq_hz = 55e3

    logger.debug("Starting endless sender")
    endless_sender.start()
    time.sleep(1)
    logger.debug("Pushing data")
    endless_sender.push_data(modulator.modulate(msg.encoded_bits))
    logger.debug("Pushed data")
    time.sleep(5)
    logger.debug("Stopping endless sender")
    endless_sender.stop()
    time.sleep(1)
    logger.debug("bye")



import os

from PyQt5.QtCore import QFileInfo
from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import QFileIconProvider

from urh import settings


class FileIconProvider(QFileIconProvider):
    def __init__(self):
        super().__init__()

    def icon(self, arg):
        if isinstance(arg, QFileInfo):
            try:
                if (
                    arg.isDir()
                    and os.path.isfile(
                        os.path.join(arg.filePath(), settings.PROJECT_FILE)
                    )
                ) or (arg.isFile() and arg.fileName() == settings.PROJECT_FILE):
                    return QIcon(":/icons/icons/appicon.png")
            except:
                # In some environments (e.g. docker) there tend to be encoding errors
                pass

        return super().icon(arg)



from PyQt5.QtCore import Qt, QPointF
from PyQt5.QtGui import QColor, QPen, QTransform
from PyQt5.QtWidgets import QGraphicsRectItem


class Selection(QGraphicsRectItem):
    def __init__(self, *args, fillcolor, opacity, parent=None):
        if len(args) == 0:
            super().__init__(parent)
        elif len(args) == 1:
            super().__init__(args[0], parent)
        elif len(args) == 4:
            x0, y0, w, h = args
            super().__init__(x0, y0, w, h, parent)

        self.finished = False
        self.selected_edge = None  # type: int
        self.resizing = False

        self.setBrush(fillcolor)
        self.setPen(QPen(QColor(Qt.transparent), 0))
        self.setOpacity(opacity)

    @property
    def is_empty(self) -> bool:
        raise NotImplementedError("Overwrite in subclass")

    @property
    def width(self):
        return self.rect().width()

    @width.setter
    def width(self, value):
        if value == self.width:
            return

        r = self.rect()
        r.setWidth(value)
        self.setRect(r)

    @property
    def height(self):
        return self.rect().height()

    @height.setter
    def height(self, value):
        if value == self.height:
            return

        r = self.rect()
        r.setHeight(value)
        self.setRect(r)

    @property
    def x(self):
        return self.rect().x()

    def setX(self, p_float):
        if p_float == self.x:
            return

        r = self.rect()
        r.setX(p_float)
        self.setRect(r)

    @property
    def y(self):
        return self.rect().y()

    def setY(self, p_float):
        if p_float == self.y:
            return

        r = self.rect()
        r.setY(p_float)
        self.setRect(r)

    @property
    def start(self):
        raise NotImplementedError("Overwrite in subclass")

    @start.setter
    def start(self, value):
        raise NotImplementedError("Overwrite in subclass")

    @property
    def end(self):
        raise NotImplementedError("Overwrite in subclass")

    @end.setter
    def end(self, value):
        raise NotImplementedError("Overwrite in subclass")

    def _get_selected_edge(
        self, pos: QPointF, transform: QTransform, horizontal_selection: bool
    ):
        x1, x2 = self.x, self.x + self.width
        y1, y2 = self.y, self.y + self.height
        x, y = pos.x(), pos.y()

        spacing = 5
        spacing /= transform.m11() if horizontal_selection else transform.m22()

        if horizontal_selection:
            x1a, x1b = x1 - spacing, x1 + spacing
            y1a, y1b = y1, y2
            x2a, x2b = x2 - spacing, x2 + spacing
            y2a, y2b = y1, y2
        else:
            x1a, x1b, x2a, x2b = x1, x2, x1, x2
            y1a, y1b = min(y1 - spacing, y1 + spacing), max(y1 - spacing, y1 + spacing)
            y2a, y2b = min(y2 - spacing, y2 + spacing), max(y2 - spacing, y2 + spacing)

        if x1a < x < x1b and y1a < y < y1b:
            self.selected_edge = 0
            return 0

        if x2a < x < x2b and y2a < y < y2b:
            self.selected_edge = 1
            return 1

        self.selected_edge = None
        return None

    def clear(self):
        self.resizing = False
        self.selected_edge = None
        self.finished = False



from PyQt5.QtCore import QRectF
from PyQt5.QtWidgets import QGraphicsTextItem

from urh.simulator.GraphicsItem import GraphicsItem
from urh.simulator.SimulatorCounterAction import SimulatorCounterAction
from urh.simulator.SimulatorGotoAction import SimulatorGotoAction
from urh.simulator.SimulatorItem import SimulatorItem
from urh.simulator.SimulatorSleepAction import SimulatorSleepAction
from urh.simulator.SimulatorTriggerCommandAction import SimulatorTriggerCommandAction


class ActionItem(GraphicsItem):
    def __init__(self, model_item: SimulatorItem, parent=None):
        super().__init__(model_item=model_item, parent=parent)

        self.setFlag(QGraphicsTextItem.ItemIsPanel, True)

        self.text = QGraphicsTextItem(self)
        self.text.setFont(self.font)

    def update_flags(self):
        if self.scene().mode == 0:
            self.set_flags(
                is_selectable=True,
                is_movable=True,
                accept_hover_events=True,
                accept_drops=True,
            )

    def update_position(self, x_pos, y_pos):
        self.setPos(x_pos, y_pos)
        start_x = (self.scene().items_width() - self.labels_width()) / 2
        self.number.setPos(start_x, 0)
        start_x += self.number.boundingRect().width()
        self.text.setPos(start_x, 0)

        width = self.scene().items_width()
        self.prepareGeometryChange()
        self.bounding_rect = QRectF(
            0, 0, width, self.childrenBoundingRect().height() + 5
        )

    def labels_width(self):
        width = self.number.boundingRect().width()
        width += self.text.boundingRect().width()
        return width


class GotoActionItem(ActionItem):
    def __init__(self, model_item: SimulatorGotoAction, parent=None):
        assert isinstance(model_item, SimulatorGotoAction)
        super().__init__(model_item=model_item, parent=parent)

    def refresh(self):
        text = "[Goto: "
        text += (
            "..."
            if self.model_item.goto_target is None
            else self.model_item.goto_target
        )
        text += "]"
        self.text.setPlainText(text)


class TriggerCommandActionItem(ActionItem):
    def __init__(self, model_item: SimulatorTriggerCommandAction, parent=None):
        super().__init__(model_item=model_item, parent=parent)
        self.text.setPlainText("[Trigger command]")


class SleepActionItem(ActionItem):
    def __init__(self, model_item: SimulatorSleepAction, parent=None):
        super().__init__(model_item=model_item, parent=parent)
        self.text.setPlainText("[" + model_item.caption + "]")

    def refresh(self):
        self.text.setPlainText("[" + self.model_item.caption + "]")


class CounterActionItem(ActionItem):
    def __init__(self, model_item: SimulatorCounterAction, parent=None):
        super().__init__(model_item=model_item, parent=parent)
        self.text.setPlainText("[Counter]")



import time
from multiprocessing import Process, Value

from urh import settings
from urh.signalprocessing.Modulator import Modulator
from urh.util.Logger import logger
from urh.util.RingBuffer import RingBuffer


class ContinuousModulator(object):
    """
    This class is used in continuous sending mode.
    You pass a list of messages and modulators to it, and it takes care of modulating the messages sequentially.
    This avoids running out of RAM for large amounts of messages.
    """

    WAIT_TIMEOUT = 0.1

    def __init__(self, messages, modulators, num_repeats=-1):
        """

        :type messages: list of Message
        :type modulators: list of Modulator
        """
        self.messages = messages
        self.modulators = modulators
        self.num_repeats = num_repeats  # -1 or 0 = infinite

        self.ring_buffer = RingBuffer(
            int(settings.CONTINUOUS_BUFFER_SIZE_MB * 1e6) // 8,
            dtype=Modulator.get_dtype(),
        )

        self.current_message_index = Value("L", 0)

        self.abort = Value("i", 0)
        self.process = Process(
            target=self.modulate_continuously, args=(self.num_repeats,), daemon=True
        )

    @property
    def is_running(self):
        return self.process.is_alive()

    def start(self):
        self.abort.value = 0
        try:
            self.process = Process(
                target=self.modulate_continuously, args=(self.num_repeats,), daemon=True
            )
            self.process.start()
        except RuntimeError as e:
            logger.exception(e)

    def stop(self, clear_buffer=True):
        self.abort.value = 1

        if self.process.is_alive():
            try:
                self.process.join(1.5)
            except RuntimeError as e:
                logger.exception(e)
                self.process.terminate()

        if clear_buffer:
            self.ring_buffer.clear()

        logger.debug("Stopped continuous modulation")

    def modulate_continuously(self, num_repeats):
        rng = (
            iter(int, 1) if num_repeats <= 0 else range(0, num_repeats)
        )  # <= 0 = forever
        for _ in rng:
            if self.abort.value:
                return

            start = self.current_message_index.value

            for i in range(start, len(self.messages)):
                if self.abort.value:
                    return

                message = self.messages[i]
                self.current_message_index.value = i
                modulator = self.modulators[message.modulator_index]  # type: Modulator
                modulated = modulator.modulate(
                    start=0, data=message.encoded_bits, pause=message.pause
                )
                while not self.ring_buffer.will_fit(len(modulated)):
                    if self.abort.value:
                        return

                    # Wait till there is space in buffer
                    time.sleep(self.WAIT_TIMEOUT)

                self.ring_buffer.push(modulated)

            self.current_message_index.value = 0



import os
import shutil
import tarfile
import tempfile
import zipfile

import numpy as np
from PyQt5.QtCore import QDir
from PyQt5.QtWidgets import QFileDialog, QMessageBox

from urh.signalprocessing.IQArray import IQArray

archives = {}
""":type: dict of [str, str]
   :param: archives[extracted_filename] = filename"""

RECENT_PATH = QDir.homePath()

SIGNAL_FILE_EXTENSIONS_BY_TYPE = {
    np.int8: ".complex16s",
    np.uint8: ".complex16u",
    np.int16: ".complex32s",
    np.uint16: ".complex32u",
    np.float32: ".complex",
    np.complex64: ".complex",
}

SIGNAL_NAME_FILTERS_BY_TYPE = {
    np.int8: "Complex16 signed (*.complex16s *.cs8)",
    np.uint8: "Complex16 unsigned (*.complex16u *.cu8)",
    np.uint16: "Complex32 unsigned (*.complex32u *.cu16)",
    np.int16: "Complex32 signed (*.complex32s *.cs16)",
    np.float32: "Complex (*.complex)",
    np.complex64: "Complex (*.complex)",
}

EVERYTHING_FILE_FILTER = "All Files (*)"

SIGNAL_NAME_FILTERS = list(sorted(set(SIGNAL_NAME_FILTERS_BY_TYPE.values())))

COMPRESSED_COMPLEX_FILE_FILTER = "Compressed Complex File (*.coco)"
WAV_FILE_FILTER = "Waveform Audio File Format (*.wav *.wave)"
PROTOCOL_FILE_FILTER = "Protocol (*.proto.xml *.proto)"
BINARY_PROTOCOL_FILE_FILTER = "Binary Protocol (*.bin)"
WIRESHARK_FILE_FILTER = "Wireshark File (*.pcapng)"
PLAIN_BITS_FILE_FILTER = "Plain Bits (*.txt)"
FUZZING_FILE_FILTER = "Fuzzing Profile (*.fuzz.xml *.fuzz)"
SIMULATOR_FILE_FILTER = "Simulator Profile (*.sim.xml *.sim)"
TAR_FILE_FILTER = "Tar Archive (*.tar *.tar.gz *.tar.bz2)"
ZIP_FILE_FILTER = "Zip Archive (*.zip)"
SUB_FILE_FILTER = "Flipper Zero SubGHz RAW (*.sub)"


def __get__name_filter_for_signals() -> str:
    return ";;".join(
        [EVERYTHING_FILE_FILTER]
        + SIGNAL_NAME_FILTERS
        + [COMPRESSED_COMPLEX_FILE_FILTER, WAV_FILE_FILTER, SUB_FILE_FILTER]
    )


def get_open_dialog(
    directory_mode=False, parent=None, name_filter="full"
) -> QFileDialog:
    dialog = QFileDialog(parent=parent, directory=RECENT_PATH)

    if directory_mode:
        dialog.setFileMode(QFileDialog.Directory)
        dialog.setWindowTitle("Open Folder")
    else:
        dialog.setFileMode(QFileDialog.ExistingFiles)
        dialog.setWindowTitle("Open Files")
        if name_filter == "full":
            name_filter = (
                __get__name_filter_for_signals()
                + ";;"
                + ";;".join(
                    [
                        PROTOCOL_FILE_FILTER,
                        BINARY_PROTOCOL_FILE_FILTER,
                        PLAIN_BITS_FILE_FILTER,
                        FUZZING_FILE_FILTER,
                        SIMULATOR_FILE_FILTER,
                        TAR_FILE_FILTER,
                        ZIP_FILE_FILTER,
                    ]
                )
            )
        elif name_filter == "signals_only":
            name_filter = __get__name_filter_for_signals()
        elif name_filter == "proto":
            name_filter = ";;".join([PROTOCOL_FILE_FILTER, BINARY_PROTOCOL_FILE_FILTER])
        elif name_filter == "fuzz":
            name_filter = FUZZING_FILE_FILTER
        elif name_filter == "simulator":
            name_filter = SIMULATOR_FILE_FILTER

        dialog.setNameFilter(name_filter)

    return dialog


def ask_save_file_name(
    initial_name: str, caption="Save signal", selected_name_filter=None
):
    global RECENT_PATH
    if caption == "Save signal":
        name_filter = __get__name_filter_for_signals()
    elif caption == "Save fuzzing profile":
        name_filter = FUZZING_FILE_FILTER
    elif caption == "Save encoding":
        name_filter = ""
    elif caption == "Save simulator profile":
        name_filter = SIMULATOR_FILE_FILTER
    elif caption == "Export spectrogram":
        name_filter = "Frequency Time (*.ft);;Frequency Time Amplitude (*.fta)"
    elif caption == "Save protocol":
        name_filter = ";;".join(
            [PROTOCOL_FILE_FILTER, BINARY_PROTOCOL_FILE_FILTER, WIRESHARK_FILE_FILTER]
        )
    elif caption == "Export demodulated":
        name_filter = ";;".join([WAV_FILE_FILTER, SUB_FILE_FILTER])
    else:
        name_filter = EVERYTHING_FILE_FILTER

    filename = None
    dialog = QFileDialog(directory=RECENT_PATH, caption=caption, filter=name_filter)
    dialog.setFileMode(QFileDialog.AnyFile)
    dialog.setLabelText(QFileDialog.Accept, "Save")
    dialog.setAcceptMode(QFileDialog.AcceptSave)

    if selected_name_filter is not None:
        dialog.selectNameFilter(selected_name_filter)

    dialog.selectFile(initial_name)

    if dialog.exec():
        filename = dialog.selectedFiles()[0]

    if filename:
        RECENT_PATH = os.path.split(filename)[0]

    return filename


def ask_signal_file_name_and_save(
    signal_name: str, data, sample_rate=1e6, wav_only=False, parent=None
) -> str:
    if wav_only:
        if not signal_name.endswith(".wav") and not signal_name.endswith(".wave"):
            signal_name += ".wav"
        selected_name_filter = WAV_FILE_FILTER
    else:
        if not any(
            signal_name.endswith(e) for e in SIGNAL_NAME_FILTERS_BY_TYPE.values()
        ):
            try:
                dtype = next(
                    d for d in SIGNAL_FILE_EXTENSIONS_BY_TYPE.keys() if d == data.dtype
                )
                signal_name += SIGNAL_FILE_EXTENSIONS_BY_TYPE[dtype]
                selected_name_filter = SIGNAL_NAME_FILTERS_BY_TYPE[dtype]
            except StopIteration:
                selected_name_filter = None
        else:
            selected_name_filter = None

    filename = ask_save_file_name(
        signal_name, selected_name_filter=selected_name_filter
    )

    if filename:
        try:
            save_data(data, filename, sample_rate=sample_rate)
        except Exception as e:
            QMessageBox.critical(parent, "Error saving signal", e.args[0])
            filename = None
    else:
        filename = None

    return filename


def save_data(data, filename: str, sample_rate=1e6, num_channels=2):
    if not isinstance(data, IQArray):
        data = IQArray(data)

    if filename.endswith(".wav"):
        data.export_to_wav(filename, num_channels, sample_rate)
    elif filename.endswith(".coco"):
        data.save_compressed(filename)
    elif filename.endswith(".sub"):
        data.export_to_sub(filename, 433920000, "FuriHalSubGhzPresetOok650Async")
    else:
        data.tofile(filename)

    if filename in archives.keys():
        archive = archives[filename]
        if archive.endswith("zip"):
            rewrite_zip(archive)
        elif (
            archive.endswith("tar") or archive.endswith("bz2") or archive.endswith("gz")
        ):
            rewrite_tar(archive)


def save_signal(signal):
    save_data(signal.iq_array.data, signal.filename, signal.sample_rate)


def rewrite_zip(zip_name):
    tempdir = tempfile.mkdtemp()
    try:
        temp_name = os.path.join(tempdir, "new.zip")
        files_in_archive = [f for f in archives.keys() if archives[f] == zip_name]
        with zipfile.ZipFile(temp_name, "w") as zip_write:
            for filename in files_in_archive:
                zip_write.write(filename)
        shutil.move(temp_name, zip_name)
    finally:
        shutil.rmtree(tempdir)


def rewrite_tar(tar_name: str):
    tempdir = tempfile.mkdtemp()
    compression = ""
    if tar_name.endswith("gz"):
        compression = "gz"
    elif tar_name.endswith("bz2"):
        compression = "bz2"
    try:
        ext = "" if len(compression) == 0 else "." + compression
        temp_name = os.path.join(tempdir, "new.tar" + ext)
        files_in_archive = [f for f in archives.keys() if archives[f] == tar_name]
        with tarfile.open(temp_name, "w:" + compression) as tar_write:
            for file in files_in_archive:
                tar_write.add(file)
        shutil.move(temp_name, tar_name)
    finally:
        shutil.rmtree(tempdir)


def uncompress_archives(file_names, temp_dir):
    """
    Extract each archive from the list of filenames.
    Normal files stay untouched.
    Add all files to the Recent Files.
    :type file_names: list of str
    :type temp_dir: str
    :rtype: list of str
    """
    result = []
    for filename in file_names:
        if (
            filename.endswith(".tar")
            or filename.endswith(".tar.gz")
            or filename.endswith(".tar.bz2")
        ):
            obj = tarfile.open(filename, "r")
            extracted_file_names = []
            for j, member in enumerate(obj.getmembers()):
                obj.extract(member, temp_dir)
                extracted_filename = os.path.join(temp_dir, obj.getnames()[j])
                extracted_file_names.append(extracted_filename)
                archives[extracted_filename] = filename
            result.extend(extracted_file_names[:])
        elif filename.endswith(".zip"):
            obj = zipfile.ZipFile(filename)
            extracted_file_names = []
            for j, info in enumerate(obj.infolist()):
                obj.extract(info, path=temp_dir)
                extracted_filename = os.path.join(temp_dir, obj.namelist()[j])
                extracted_file_names.append(extracted_filename)
                archives[extracted_filename] = filename
            result.extend(extracted_file_names[:])
        else:
            result.append(filename)

    return result


def get_directory():
    directory = QFileDialog.getExistingDirectory(
        None,
        "Choose Directory",
        QDir.homePath(),
        QFileDialog.ShowDirsOnly | QFileDialog.DontResolveSymlinks,
    )
    return directory
